import numpy as np
import pandas as pd
from six import string_types

from triage.component.catwalk.exceptions import BaselineFeatureNotInMatrix

REQUIRED_KEYS = frozenset(['feature_name', 'operator', 'threshold'])


class SimpleThresholder(object):
    """ The simple thresholder applies a set of predetermined logical rules to a
    test matrix to classify entities. By default, it will classify entities as 1
    if they satisfy any of the rules. When 'and' is set as the logical_operator,
    it will classify entities as 1 only if they pass *all* of the rules.

    Rules are passed as either strings in the format 'x1 > 5' or dictionaries in
    the format {feature_name: 'x1', operator: '>', threshold: 5}. The
    feature_name, operator, and threshold keys are required. Eventually, this
    class may be abstracted into a BaseThreshold class and more complicated
    thresholders could be built around new keys in the dictionaries (e.g., by
    specifying scores that could be applied (and possibly summed) to entities
    satisfying rules) or by an alternative dictionary format that specifies
    more complicated structures for applying rules (for example:
        {
            or: [
                {or: [{}, {}]},
                {and: [{}, {}]}
            ]
        }
    where rules and operators that combine them can be nested).
    """
    def __init__(self, rules, logical_operator='or'):
        self.rules = rules
        self.logical_operator = logical_operator
        self.feature_importances_ = None
        self.rule_combination_method = self.lookup_rule_combination_method(logical_operator)

    @property
    def rules(self):
        return vars(self)['rules']

    @rules.setter
    def rules(self, rules):
        """ Validates the rules passed by the user and converts them to the
        internal representation. Can be used to validate rules before running an
        experiment.

        1. If rules are not a list, make them a list.
        2. If rules are strings, convert them to dictionaries.
        3. If dictionaries or strings are not in a supported format, raise
           helpful exceptions.
        """
        if not isinstance(rules, list):
            rules = [rules]

        converted_rules = []
        for rule in rules:
            if isinstance(rule, string_types):
                converted_rules.append(self._convert_string_rule_to_dict(rule))
            else:
                if not isinstance(rule, dict):
                    raise ValueError((
                        'Rule "{rule}" is not of a supported type (string or '
                        'dict).'.format(rule=rule)
                    ))
                if not rule.keys() >= REQUIRED_KEYS:
                    raise ValueError((
                        'Rule "{rule}" missing one or more required keys '
                        '({required_keys}).'.format(
                            rule=rule,
                            required_keys=REQUIRED_KEYS
                        )
                    ))
                rule['operator'] = self._validate_operator(rule['operator'], rule)
                converted_rules.append(rule)

        vars(self)['rules'] = converted_rules

    @property
    def all_feature_names(self):
        return [rule['feature_name'] for rule in self.rules]

    def lookup_rule_combination_method(self, logical_operator):
        """ Convert 'and' to 'all' and 'or' to 'any' for interacting with
        pandas DataFrames.
        """
        rule_combination_method_lookup = {
            'or': 'any',
            'and': 'all'
        }
        return rule_combination_method_lookup[logical_operator]

    def _validate_operator(self, operator_string, rule):
        """ Convert the user-passed operator into the the name of the apprpriate
        pandas method.
        """
        operator_lookup = {
            '>': 'gt',
            '>=': 'ge',
            '<': 'lt',
            '<=': 'le',
            '==': 'equals'
        }
        try:
            operator = operator_lookup[operator_string]
        except KeyError:
            raise ValueError((
                'Operator "{operator}" extracted from rule "{rule}" is not a '
                'supported operator ({supported_operators}).'.format(
                    operator=operator_string,
                    rule=rule,
                    supported_operators=operator_lookup.keys()
                )
            ))

        return operator

    def _convert_string_rule_to_dict(self, rule):
        """ Converts a string rule into a dict, raising helpful exceptions if it
        cannot be parsed.
        """
        components = rule.rsplit(' ', 2)

        if len(components) < 3:
            raise ValueError((
                '{required_keys} could not be parsed from rule "{rule}". Are '
                'they all present and separated by spaces?'.format(
                    required_keys=REQUIRED_KEYS,
                    rule=rule
                )
            ))

        try:
            threshold = int(components[2])
        except ValueError:
            raise ValueError((
                'Threshold "{threshold}" parsed from rule "{rule}" is not an '
                'int.'.format(threshold=components[2], rule=rule)
            ))

        operator = self._validate_operator(components[1], rule)

        return {
            'feature_name': components[0],
            'operator': operator,
            'threshold': threshold
        }

    def _set_feature_importances_(self, x):
        """ Assigns feature importances following the rule: 1 for the features
        we are thresholding on, 0 for all other features.
        """
        feature_importances = [0] * len(x.columns)
        for feature_name in self.all_feature_names:
            try:
                position = x.columns.get_loc(feature_name)
            except KeyError:
                raise BaselineFeatureNotInMatrix((
                    'Rules refer to a feature ({feature_name}) not included in '
                    'the training matrix!'.format(feature_name=feature_name)
                ))
            feature_importances[position] = 1
        self.feature_importances_ = np.array(feature_importances)

    def fit(self, x, y):
        """ Set feature importances and return self.
        """
        self._set_feature_importances_(x)
        return self

    def predict_proba(self, x):
        """ Assign 1 for entities that meet the rules and 0 for those that do not.
        """
        rule_evaluations_list = []
        for rule in self.rules:
            rule_evaluations_list.append(
                getattr(x[rule['feature_name']], rule['operator'])(rule['threshold'])
            )
        rule_evaluations_dataframe = pd.concat(rule_evaluations_list, axis=1)
        scores = getattr(rule_evaluations_dataframe, self.rule_combination_method)(axis=1)
        scores = list(scores.astype(int))

        # format it like sklearn output and return
        return np.array([scores, scores]).transpose()



import numpy as np

import warnings

import pytest

from triage.component.catwalk.estimators.transformers import CutOff
from triage.component.catwalk.estimators.classifiers import ScaledLogisticRegression

from sklearn import linear_model

from sklearn import datasets
from sklearn import preprocessing
from sklearn.pipeline import Pipeline
from sklearn.model_selection import train_test_split

@pytest.fixture
def data():
    dataset = datasets.load_breast_cancer()
    X = dataset.data
    y = dataset.target

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = 0.3, random_state=12345)

    return {'X_train':X_train, 'X_test':X_test, 'y_train':y_train, 'y_test':y_test}

def test_cutoff_warning():
    X_data = [1, 2, 0.5, 0.7, 100, -1, -23, 0]

    cutoff = CutOff()

    with pytest.warns(UserWarning):
       cutoff.fit_transform(X_data)


def test_cutoff_transformer():
    cutoff = CutOff()

    X_data = [1, 2, 0.5, 0.7, 100, -1, -23, 0]

    assert np.all(cutoff.fit_transform(X_data) == [1, 1, 0.5, 0.7, 1, 0, 0, 0])

def test_cutoff_inside_a_pipeline(data):
    minmax_scaler = preprocessing.MinMaxScaler()
    dsapp_cutoff = CutOff()

    pipeline =Pipeline([
        ('minmax_scaler',minmax_scaler),
        ('dsapp_cutoff', dsapp_cutoff)
    ])

    pipeline.fit(data['X_train'], data['y_train'])

    X_fake_new_data = data['X_test'][-1,:].reshape(1,-1) + 0.5

    mms = preprocessing.MinMaxScaler().fit(data['X_train'])

    assert np.all(( mms.transform(X_fake_new_data) > 1  ) == (pipeline.transform(X_fake_new_data) == 1))

def test_dsapp_lr(data):
    dsapp_lr = ScaledLogisticRegression()
    dsapp_lr.fit(data['X_train'], data['y_train'])

    minmax_scaler = preprocessing.MinMaxScaler()
    dsapp_cutoff = CutOff()
    lr = linear_model.LogisticRegression()

    pipeline =Pipeline([
        ('minmax_scaler',minmax_scaler),
        ('dsapp_cutoff', dsapp_cutoff),
        ('lr', lr)
    ])

    pipeline.fit(data['X_train'], data['y_train'])

    assert np.all(dsapp_lr.predict(data['X_test']) == pipeline.predict(data['X_test']))



# -*- coding: utf-8 -*-
"""test_spacetime

Unit tests for `collate.spacetime` module.

"""
from datetime import date
from itertools import product

import pytest
import sqlalchemy
import testing.postgresql

from triage.component.collate import Aggregate, SpacetimeAggregation


events_data = [
    # entity id, event_date, outcome
    [1, date(2014, 1, 1), True],
    [1, date(2014, 11, 10), False],
    [1, date(2015, 1, 1), False],
    [1, date(2015, 11, 10), True],
    [2, date(2013, 6, 8), True],
    [2, date(2014, 6, 8), False],
    [3, date(2014, 3, 3), False],
    [3, date(2014, 7, 24), False],
    [3, date(2015, 3, 3), True],
    [3, date(2015, 7, 24), False],
    [4, date(2015, 12, 13), False],
    [4, date(2016, 12, 13), True],
]

# distinct entity_id, event_date pairs
state_data = sorted(list(product(
    set([l[0] for l in events_data]),
    set([l[1] for l in events_data] + [date(2016, 1, 1)])
)))


def test_basic_spacetime():
    with testing.postgresql.Postgresql() as psql:
        engine = sqlalchemy.create_engine(psql.url())
        engine.execute(
            'create table events (entity_id int, event_date date, outcome bool)'
        )
        for event in events_data:
            engine.execute(
                'insert into events values (%s, %s, %s::bool)',
                event
            )

        engine.execute(
            'create table states (entity_id int, as_of_date date)'
        )
        for state in state_data:
            engine.execute(
                'insert into states values (%s, %s)',
                state
            )

        agg = Aggregate('outcome::int', ['sum','avg'], {
            "coltype": "aggregate",
            "avg": {"type": "mean"},
            "sum": {"type": "constant", "value": 3},
            "max": {"type": "zero"}
        })
        st = SpacetimeAggregation(
                aggregates = [agg],
                from_obj = 'events',
                groups = ['entity_id'],
                intervals = ['1y', '2y', 'all'],
                dates = ['2016-01-01', '2015-01-01'],
                state_table = 'states',
                state_group = 'entity_id',
                date_column = 'event_date',
                output_date_column = 'as_of_date'
            )

        st.execute(engine.connect())

        r = engine.execute('select * from events_entity_id order by entity_id, as_of_date')
        rows = [x for x in r]
        assert rows[0]['entity_id'] == 1
        assert rows[0]['as_of_date'] == date(2015, 1, 1)
        assert rows[0]['events_entity_id_1y_outcome::int_sum'] == 1
        assert rows[0]['events_entity_id_1y_outcome::int_avg'] == 0.5
        assert rows[0]['events_entity_id_2y_outcome::int_sum'] == 1
        assert rows[0]['events_entity_id_2y_outcome::int_avg'] == 0.5
        assert rows[0]['events_entity_id_all_outcome::int_sum'] == 1
        assert rows[0]['events_entity_id_all_outcome::int_avg'] == 0.5
        assert rows[1]['entity_id'] == 1
        assert rows[1]['as_of_date'] == date(2016, 1, 1)
        assert rows[1]['events_entity_id_1y_outcome::int_sum'] == 1
        assert rows[1]['events_entity_id_1y_outcome::int_avg'] == 0.5
        assert rows[1]['events_entity_id_2y_outcome::int_sum'] == 2
        assert rows[1]['events_entity_id_2y_outcome::int_avg'] == 0.5
        assert rows[1]['events_entity_id_all_outcome::int_sum'] == 2
        assert rows[1]['events_entity_id_all_outcome::int_avg'] == 0.5

        assert rows[2]['entity_id'] == 2
        assert rows[2]['as_of_date'] == date(2015, 1, 1)
        assert rows[2]['events_entity_id_1y_outcome::int_sum'] == 0
        assert rows[2]['events_entity_id_1y_outcome::int_avg'] == 0
        assert rows[2]['events_entity_id_2y_outcome::int_sum'] == 1
        assert rows[2]['events_entity_id_2y_outcome::int_avg'] == 0.5
        assert rows[2]['events_entity_id_all_outcome::int_sum'] == 1
        assert rows[2]['events_entity_id_all_outcome::int_avg'] == 0.5
        assert rows[3]['entity_id'] == 2
        assert rows[3]['as_of_date'] == date(2016, 1, 1)
        assert rows[3]['events_entity_id_1y_outcome::int_sum'] == None
        assert rows[3]['events_entity_id_1y_outcome::int_avg'] == None
        assert rows[3]['events_entity_id_2y_outcome::int_sum'] == 0
        assert rows[3]['events_entity_id_2y_outcome::int_avg'] == 0
        assert rows[3]['events_entity_id_all_outcome::int_sum'] == 1
        assert rows[3]['events_entity_id_all_outcome::int_avg'] == 0.5

        assert rows[4]['entity_id'] == 3
        assert rows[4]['as_of_date'] == date(2015, 1, 1)
        assert rows[4]['events_entity_id_1y_outcome::int_sum'] == 0
        assert rows[4]['events_entity_id_1y_outcome::int_avg'] == 0
        assert rows[4]['events_entity_id_2y_outcome::int_sum'] == 0
        assert rows[4]['events_entity_id_2y_outcome::int_avg'] == 0
        assert rows[4]['events_entity_id_all_outcome::int_sum'] == 0
        assert rows[4]['events_entity_id_all_outcome::int_avg'] == 0
        assert rows[5]['entity_id'] == 3
        assert rows[5]['as_of_date'] == date(2016, 1, 1)
        assert rows[5]['events_entity_id_1y_outcome::int_sum'] == 1
        assert rows[5]['events_entity_id_1y_outcome::int_avg'] == 0.5
        assert rows[5]['events_entity_id_2y_outcome::int_sum'] == 1
        assert rows[5]['events_entity_id_2y_outcome::int_avg'] == 0.25
        assert rows[5]['events_entity_id_all_outcome::int_sum'] == 1
        assert rows[5]['events_entity_id_all_outcome::int_avg'] == 0.25

        assert rows[6]['entity_id'] == 4
        # rows[6]['date'] == date(2015, 1, 1) is skipped due to no data!
        assert rows[6]['as_of_date'] == date(2016, 1, 1)
        assert rows[6]['events_entity_id_1y_outcome::int_sum'] == 0
        assert rows[6]['events_entity_id_1y_outcome::int_avg'] == 0
        assert rows[6]['events_entity_id_2y_outcome::int_sum'] == 0
        assert rows[6]['events_entity_id_2y_outcome::int_avg'] == 0
        assert rows[6]['events_entity_id_all_outcome::int_sum'] == 0
        assert rows[6]['events_entity_id_all_outcome::int_avg'] == 0
        assert len(rows) == 7

        # check some imputation results
        r = engine.execute('select * from events_aggregation_imputed order by entity_id, as_of_date')
        rows = [x for x in r]
        assert rows[6]['entity_id'] == 4
        assert rows[6]['as_of_date'] == date(2015, 1, 1)
        assert rows[6]['events_entity_id_1y_outcome::int_sum'] == 3
        assert rows[6]['events_entity_id_1y_outcome::int_sum_imp'] == 1
        assert round(float(rows[6]['events_entity_id_1y_outcome::int_avg']), 4) == 0.1667
        assert rows[6]['events_entity_id_1y_outcome::int_avg_imp'] == 1
        assert rows[6]['events_entity_id_2y_outcome::int_sum'] == 3
        assert rows[6]['events_entity_id_2y_outcome::int_sum_imp'] == 1
        assert round(float(rows[6]['events_entity_id_2y_outcome::int_avg']), 4) == 0.3333
        assert rows[6]['events_entity_id_2y_outcome::int_avg_imp'] == 1
        assert rows[6]['events_entity_id_all_outcome::int_sum'] == 3
        assert rows[6]['events_entity_id_all_outcome::int_sum_imp'] == 1
        assert round(float(rows[6]['events_entity_id_all_outcome::int_avg']), 4) == 0.3333
        assert rows[6]['events_entity_id_all_outcome::int_avg_imp'] == 1
        assert rows[7]['entity_id'] == 4
        assert rows[7]['as_of_date'] == date(2016, 1, 1)
        assert rows[7]['events_entity_id_1y_outcome::int_sum'] == 0
        assert rows[7]['events_entity_id_1y_outcome::int_sum_imp'] == 0
        assert rows[7]['events_entity_id_1y_outcome::int_avg'] == 0
        assert rows[7]['events_entity_id_1y_outcome::int_avg_imp'] == 0
        assert rows[7]['events_entity_id_2y_outcome::int_sum'] == 0
        assert rows[7]['events_entity_id_2y_outcome::int_sum_imp'] == 0
        assert rows[7]['events_entity_id_2y_outcome::int_avg'] == 0
        assert rows[7]['events_entity_id_2y_outcome::int_avg_imp'] == 0
        assert rows[7]['events_entity_id_all_outcome::int_sum'] == 0
        assert rows[7]['events_entity_id_all_outcome::int_sum_imp'] == 0
        assert rows[7]['events_entity_id_all_outcome::int_avg'] == 0
        assert rows[7]['events_entity_id_all_outcome::int_avg_imp'] == 0
        assert len(rows) == 8


def test_input_min_date():
    with testing.postgresql.Postgresql() as psql:
        engine = sqlalchemy.create_engine(psql.url())
        engine.execute(
            'create table events (entity_id int, date date, outcome bool)'
        )
        for event in events_data:
            engine.execute(
                'insert into events values (%s, %s, %s::bool)',
                event
            )

        engine.execute(
            'create table states (entity_id int, date date)'
        )
        for state in state_data:
            engine.execute(
                'insert into states values (%s, %s)',
                state
            )

        agg = Aggregate('outcome::int', ['sum','avg'], {
            "coltype": "aggregate",
            "avg": {"type": "mean"},
            "sum": {"type": "constant", "value": 3},
            "max": {"type": "zero"}
        })
        st = SpacetimeAggregation(
            aggregates = [agg],
            from_obj = 'events',
            groups = ['entity_id'],
            intervals = ['all'],
            dates = ['2016-01-01'],
            state_table = 'states',
            state_group = 'entity_id',
            date_column = '"date"',
            input_min_date = '2015-11-10'
        )

        st.execute(engine.connect())

        r = engine.execute('select * from events_entity_id order by entity_id')
        rows = [x for x in r]

        assert rows[0]['entity_id'] == 1
        assert rows[0]['date'] == date(2016, 1, 1)
        assert rows[0]['events_entity_id_all_outcome::int_sum'] == 1
        assert rows[0]['events_entity_id_all_outcome::int_avg'] == 1
        assert rows[1]['entity_id'] == 4
        assert rows[1]['date'] == date(2016, 1, 1)
        assert rows[1]['events_entity_id_all_outcome::int_sum'] == 0
        assert rows[1]['events_entity_id_all_outcome::int_avg'] == 0

        assert len(rows) == 2

        st = SpacetimeAggregation(
            aggregates = [agg],
            from_obj = 'events',
            groups = ['entity_id'],
            intervals = ['1y', 'all'],
            dates = ['2016-01-01', '2015-01-01'],
            state_table = 'states',
            state_group = 'entity_id',
            date_column = '"date"',
            input_min_date = '2014-11-10'
        )
        with pytest.raises(ValueError):
            st.validate(engine.connect())
        with pytest.raises(ValueError):
            st.execute(engine.connect())



from datetime import datetime

import pytest
import testing.postgresql
from sqlalchemy.engine import create_engine

from triage.component.architect.state_table_generators import StateTableGeneratorFromDense, StateTableGeneratorFromEntities, StateTableGeneratorFromQuery

from . import utils


def test_sparse_state_table_generator():
    input_data = [
        (5, 'permitted', datetime(2016, 1, 1), datetime(2016, 6, 1)),
        (6, 'permitted', datetime(2016, 2, 5), datetime(2016, 5, 5)),
        (1, 'injail', datetime(2014, 7, 7), datetime(2014, 7, 15)),
        (1, 'injail', datetime(2016, 3, 7), datetime(2016, 4, 2)),
    ]

    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_dense_state_table(engine, 'states', input_data)

        table_generator = StateTableGeneratorFromDense(
            db_engine=engine,
            experiment_hash='exp_hash',
            dense_state_table='states'
        )
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 2, 1),
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
            datetime(2016, 5, 1),
            datetime(2016, 6, 1),
        ]
        table_generator.generate_sparse_table(as_of_dates)
        results = [row for row in engine.execute(
            'select entity_id, as_of_date, injail, permitted from {} order by entity_id, as_of_date'.format(
                table_generator.sparse_table_name
            ))]
        expected_output = [
            # entity_id, as_of_date, injail, permitted
            (1, datetime(2016, 4, 1), True, False),
            (5, datetime(2016, 1, 1), False, True),
            (5, datetime(2016, 2, 1), False, True),
            (5, datetime(2016, 3, 1), False, True),
            (5, datetime(2016, 4, 1), False, True),
            (5, datetime(2016, 5, 1), False, True),
            (6, datetime(2016, 3, 1), False, True),
            (6, datetime(2016, 4, 1), False, True),
            (6, datetime(2016, 5, 1), False, True),
        ]
        assert results == expected_output
        utils.assert_index(engine, table_generator.sparse_table_name, 'entity_id')
        utils.assert_index(engine, table_generator.sparse_table_name, 'as_of_date')


def test_empty_dense_state_table():
    """An empty dense (input) state table produces a useful error."""
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_dense_state_table(engine, 'states', ())  # no data
        table_generator = StateTableGeneratorFromDense(
            db_engine=engine,
            experiment_hash='exp_hash',
            dense_state_table='states'
        )

        with pytest.raises(ValueError):
            table_generator.generate_sparse_table([datetime(2016, 1, 1)])

        engine.dispose()


def test_empty_sparse_state_table():
    """An empty sparse (generated) state table eagerly produces an
    error.

    (Rather than allowing execution to proceed.)

    """
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_dense_state_table(engine, 'states', (
            (5, 'permitted', datetime(2016, 1, 1), datetime(2016, 6, 1)),
            (6, 'permitted', datetime(2016, 2, 5), datetime(2016, 5, 5)),
            (1, 'injail', datetime(2014, 7, 7), datetime(2014, 7, 15)),
            (1, 'injail', datetime(2016, 3, 7), datetime(2016, 4, 2)),
        ))
        table_generator = StateTableGeneratorFromDense(
            db_engine=engine,
            experiment_hash='exp_hash',
            dense_state_table='states'
        )

        with pytest.raises(ValueError):
            # Request time outside of available intervals
            table_generator.generate_sparse_table([datetime(2015, 12, 31)])

        (state_count,) = engine.execute('''\
            select count(*) from {generator.sparse_table_name}
        '''.format(generator=table_generator)
        ).first()

        assert state_count == 0

        engine.dispose()


def test_sparse_table_generator_from_entities():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 1, 1), True),
        (3, datetime(2016, 1, 1), True),
        (5, datetime(2016, 1, 1), True),
        (5, datetime(2016, 1, 1), True),
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, 'events', input_data)
        table_generator = StateTableGeneratorFromEntities(
            entities_table='events',
            db_engine=engine,
            experiment_hash='exp_hash',
        )
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 2, 1),
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
            datetime(2016, 5, 1),
            datetime(2016, 6, 1),
        ]
        table_generator.generate_sparse_table(as_of_dates)
        expected_output = [
            (1, datetime(2016, 1, 1), True),
            (1, datetime(2016, 2, 1), True),
            (1, datetime(2016, 3, 1), True),
            (1, datetime(2016, 4, 1), True),
            (1, datetime(2016, 5, 1), True),
            (1, datetime(2016, 6, 1), True),
            (2, datetime(2016, 1, 1), True),
            (2, datetime(2016, 2, 1), True),
            (2, datetime(2016, 3, 1), True),
            (2, datetime(2016, 4, 1), True),
            (2, datetime(2016, 5, 1), True),
            (2, datetime(2016, 6, 1), True),
            (3, datetime(2016, 1, 1), True),
            (3, datetime(2016, 2, 1), True),
            (3, datetime(2016, 3, 1), True),
            (3, datetime(2016, 4, 1), True),
            (3, datetime(2016, 5, 1), True),
            (3, datetime(2016, 6, 1), True),
            (5, datetime(2016, 1, 1), True),
            (5, datetime(2016, 2, 1), True),
            (5, datetime(2016, 3, 1), True),
            (5, datetime(2016, 4, 1), True),
            (5, datetime(2016, 5, 1), True),
            (5, datetime(2016, 6, 1), True),
        ]
        results = [row for row in engine.execute(
            '''
                select entity_id, as_of_date, active from {}
                order by entity_id, as_of_date
            '''.format(
                table_generator.sparse_table_name
            )
        )]
        assert results == expected_output
        utils.assert_index(engine, table_generator.sparse_table_name, 'entity_id')
        utils.assert_index(engine, table_generator.sparse_table_name, 'as_of_date')


def test_sparse_states_from_query():
    input_data = [
        (1, datetime(2016, 1, 1), True),
        (1, datetime(2016, 4, 1), False),
        (1, datetime(2016, 3, 1), True),
        (2, datetime(2016, 1, 1), False),
        (2, datetime(2016, 1, 1), True),
        (3, datetime(2016, 1, 1), True),
        (5, datetime(2016, 3, 1), True),
        (5, datetime(2016, 4, 1), True),
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        utils.create_binary_outcome_events(engine, 'events', input_data)
        table_generator = StateTableGeneratorFromQuery(
            query="select entity_id from events where outcome_date < '{as_of_date}'::date",
            db_engine=engine,
            experiment_hash='exp_hash',
        )
        as_of_dates = [
            datetime(2016, 1, 1),
            datetime(2016, 2, 1),
            datetime(2016, 3, 1),
            datetime(2016, 4, 1),
            datetime(2016, 5, 1),
            datetime(2016, 6, 1),
        ]
        table_generator.generate_sparse_table(as_of_dates)
        expected_output = [
            (1, datetime(2016, 2, 1), True),
            (1, datetime(2016, 3, 1), True),
            (1, datetime(2016, 4, 1), True),
            (1, datetime(2016, 5, 1), True),
            (1, datetime(2016, 6, 1), True),
            (2, datetime(2016, 2, 1), True),
            (2, datetime(2016, 3, 1), True),
            (2, datetime(2016, 4, 1), True),
            (2, datetime(2016, 5, 1), True),
            (2, datetime(2016, 6, 1), True),
            (3, datetime(2016, 2, 1), True),
            (3, datetime(2016, 3, 1), True),
            (3, datetime(2016, 4, 1), True),
            (3, datetime(2016, 5, 1), True),
            (3, datetime(2016, 6, 1), True),
            (5, datetime(2016, 4, 1), True),
            (5, datetime(2016, 5, 1), True),
            (5, datetime(2016, 6, 1), True),
        ]
        results = [row for row in engine.execute(
            '''
                select entity_id, as_of_date, active from {}
                order by entity_id, as_of_date
            '''.format(
                table_generator.sparse_table_name
            )
        )]
        assert results == expected_output
        utils.assert_index(engine, table_generator.sparse_table_name, 'entity_id')
        utils.assert_index(engine, table_generator.sparse_table_name, 'as_of_date')



import csv
import logging
import math
import tempfile
from datetime import datetime

import numpy
import pandas
import postgres_copy
from sqlalchemy.orm import sessionmaker

from triage.component.results_schema import Model

from .utils import db_retry


class ModelNotFoundError(ValueError):
    pass


class Predictor(object):
    expected_matrix_ts_format = '%Y-%m-%d %H:%M:%S'

    def __init__(
        self,
        model_storage_engine,
        db_engine,
        replace=True
    ):
        """Encapsulates the task of generating predictions on an arbitrary
        dataset and storing the results

        Args:
            model_storage_engine (catwalk.storage.ModelStorageEngine)
            db_engine (sqlalchemy.engine)

        """
        self.model_storage_engine = model_storage_engine
        self.db_engine = db_engine
        self.replace = replace

    @property
    def sessionmaker(self):
        return sessionmaker(bind=self.db_engine)

    @db_retry
    def _retrieve_model_hash(self, model_id):
        """Retrieves the model hash associated with a given model id

        Args:
            model_id (int) The id of a given model in the database

        Returns: (str) the stored hash of the model
        """
        try:
            session = self.sessionmaker()
            model_hash = session.query(Model).get(model_id).model_hash
        finally:
            session.close()
        return model_hash

    @db_retry
    def load_model(self, model_id):
        """Downloads the cached model associated with a given model id

        Args:
            model_id (int) The id of a given model in the database

        Returns:
            A python object which implements .predict()
        """

        model_hash = self._retrieve_model_hash(model_id)
        logging.info('Checking for model_hash %s in store', model_hash)
        if self.model_storage_engine.exists(model_hash):
            return self.model_storage_engine.load(model_hash)

    @db_retry
    def delete_model(self, model_id):
        """Deletes the cached model associated with a given model id

        Args:
            model_id (int) The id of a given model in the database
        """
        model_hash = self._retrieve_model_hash(model_id)
        self.model_storage_engine.delete(model_hash)

    @db_retry
    def _existing_predictions(self, Prediction_obj, session, model_id, matrix_store):
        return session.query(Prediction_obj)\
            .filter_by(model_id=model_id)\
            .filter(Prediction_obj.as_of_date.in_(self._as_of_dates(matrix_store)))

    def _as_of_dates(self, matrix_store):
        matrix = matrix_store.matrix
        if 'as_of_date' in matrix.index.names:
            return matrix.index.levels[
                matrix.index.names.index('as_of_date')
            ].tolist()
        else:
            return [matrix_store.metadata['end_time']]

    @db_retry
    def _load_saved_predictions(self, existing_predictions, matrix_store):
        index = matrix_store.matrix.index
        score_lookup = {}
        for prediction in existing_predictions:
            score_lookup[(
                prediction.entity_id,
                prediction.as_of_date.date()
            )] = prediction.score
        if 'as_of_date' in index.names:
            score_iterator = (
                score_lookup[(entity_id, dt.date())]
                for (entity_id, dt) in index
            )
        else:
            as_of_date = matrix_store.metadata['end_time'].date()
            score_iterator = (score_lookup[(row, as_of_date)] for row in index)
        return numpy.fromiter(score_iterator, float)

    @db_retry
    def _write_to_db(
        self,
        model_id,
        matrix_store,
        predictions,
        labels,
        misc_db_parameters,
        Prediction_obj
    ):
        """Writes given predictions to database

        entity_ids, predictions, labels are expected to be in the same order

        Args:
            model_id (int) the id of the model associated with the given predictions
            matrix_store (catwalk.storage.MatrixStore) the matrix and metadata
            entity_ids (iterable) entity ids that predictions were made on
            predictions (iterable) predicted values
            labels (iterable) labels of prediction set (int) the id of the model
            to predict based off of
            Prediction_obj (TrainPrediction or TestPrediction) table to store predictions to

        """
        session = self.sessionmaker()
        self._existing_predictions(Prediction_obj, session, model_id, matrix_store)\
            .delete(synchronize_session=False)
        session.expire_all()
        db_objects = []
        test_label_timespan = matrix_store.metadata['label_timespan']
        logging.warning(test_label_timespan)

        if 'as_of_date' in matrix_store.matrix.index.names:
            logging.info('as_of_date found as part of matrix index, using '
                         'index for table as_of_dates')
            session.commit()
            session.close()
            with tempfile.TemporaryFile(mode='w+') as f:
                writer = csv.writer(f, quoting=csv.QUOTE_MINIMAL)
                for index, score, label in zip(
                    matrix_store.matrix.index,
                    predictions,
                    labels
                ):
                    entity_id, as_of_date = index
                    prediction = Prediction_obj(
                        model_id=int(model_id),
                        entity_id=int(entity_id),
                        as_of_date=as_of_date,
                        score=float(score),
                        label_value=int(label) if not math.isnan(label) else None,
                        matrix_uuid=matrix_store.uuid,
                        test_label_timespan=test_label_timespan,
                        **misc_db_parameters
                    )
                    writer.writerow([
                        prediction.model_id,
                        prediction.entity_id,
                        prediction.as_of_date,
                        prediction.score,
                        prediction.label_value,
                        prediction.rank_abs,
                        prediction.rank_pct,
                        prediction.matrix_uuid,
                        prediction.test_label_timespan
                    ])
                f.seek(0)
                postgres_copy.copy_from(f, Prediction_obj, self.db_engine, format='csv')
        else:
            logging.info('as_of_date not found as part of matrix index, using '
                         'matrix metadata end_time as as_of_date')
            temp_df = pandas.DataFrame({'score': predictions})
            rankings_abs = temp_df['score'].rank(method='dense', ascending=False)
            rankings_pct = temp_df['score'].rank(method='dense', ascending=False, pct=True)
            for entity_id, score, label, rank_abs, rank_pct in zip(
                matrix_store.matrix.index,
                predictions,
                labels,
                rankings_abs,
                rankings_pct
            ):
                db_objects.append(Prediction_obj(
                    model_id=int(model_id),
                    entity_id=int(entity_id),
                    as_of_date=matrix_store.metadata['end_time'],
                    score=round(float(score), 10),
                    label_value=int(label) if not math.isnan(label) else None,
                    rank_abs=int(rank_abs),
                    rank_pct=round(float(rank_pct), 10),
                    matrix_uuid=matrix_store.uuid,
                    test_label_timespan=test_label_timespan,
                    **misc_db_parameters
                ))

            session.bulk_save_objects(db_objects)
            session.commit()
            session.close()

    def predict(self, model_id, matrix_store, misc_db_parameters, train_matrix_columns):
        """Generate predictions and store them in the database

        Args:
            model_id (int) the id of the trained model to predict based off of
            matrix_store (catwalk.storage.MatrixStore) a wrapper for the
                prediction matrix and metadata
            misc_db_parameters (dict): attributes and values to add to each
                TrainPrediction or TestPrediction object in the results schema
            train_matrix_columns (list): The order of columns that the model
                was trained on

        Returns:
            (numpy.Array) the generated prediction values
        """
        #Setting the Prediction object type - TrainPrediction or TestPrediction
        prediction_obj = matrix_store.matrix_type.prediction_obj

        session = self.sessionmaker()
        if not self.replace:
            logging.info(
                'replace flag not set for model id %s, matrix %s, looking for old predictions',
                model_id,
                matrix_store.uuid
            )
            existing_predictions = self._existing_predictions(
                prediction_obj,
                session,
                model_id,
                matrix_store
            )
            index = matrix_store.matrix.index
            if existing_predictions.count() == len(index):
                logging.info(
                    'Found predictions for model id %s, matrix %s, returning saved versions',
                    model_id,
                    matrix_store.uuid
                )
                return self._load_saved_predictions(
                    existing_predictions,
                    matrix_store
                )

        model = self.load_model(model_id)
        logging.info('Loaded model %s', model_id)
        if not model:
            raise ModelNotFoundError('Model id {} not found'.format(model_id))

        # Labels are popped from matrix (IE, they are removed and returned)
        labels = matrix_store.labels()

        predictions_proba = model.predict_proba(
            matrix_store.matrix_with_sorted_columns(train_matrix_columns)
        )
        logging.info('Generated predictions for model %s, matrix %s', model_id, matrix_store.uuid)

        self._write_to_db(
            model_id,
            matrix_store,
            predictions_proba[:, 1],
            labels,
            misc_db_parameters,
            prediction_obj
        )
        logging.info(
            'Wrote predictions for model %s, matrix %s to database',
            model_id,
            matrix_store.uuid
        )
        return predictions_proba[:, 1]



"""model_group_stored_procedure

Revision ID: d0ac573eaf1a
Revises: 2446a931de7a
Create Date: 2018-06-20 17:44:27.162699

"""
from alembic import op
import os
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd0ac573eaf1a'
down_revision = '2446a931de7a'
branch_labels = None
depends_on = None


def upgrade():
    group_proc_filename = os.path.join(
        os.path.dirname(__file__),
        '../../model_group_stored_procedure.sql'
    )
    with open(group_proc_filename) as fd:
        stmt = fd.read()
        op.execute(stmt)


def downgrade():
    pass



import logging
from collections import OrderedDict

import sqlalchemy
import sqlparse

from triage.util.conf import convert_str_to_relativedelta

from triage.component.collate import (
    Aggregate,
    Categorical,
    Compare,
    SpacetimeAggregation,
)


class FeatureGenerator(object):

    def __init__(
        self,
        db_engine,
        features_schema_name,
        replace=True,
        feature_start_time=None
    ):
        """Generates aggregate features using collate

        Args:
            db_engine (sqlalchemy.db.engine)
            features_schema_name (string) Name of schema where feature
                tables should be written to
            replace (boolean, optional) Whether or not existing features
                should be replaced
            feature_start_time (string/datetime, optional) point in time before which
                should not be included in features
        """
        self.db_engine = db_engine
        self.features_schema_name = features_schema_name
        self.categorical_cache = {}
        self.replace = replace
        self.feature_start_time = feature_start_time
        self.entity_id_column = 'entity_id'

    def _validate_keys(self, aggregation_config):
        for key in ['from_obj', 'intervals', 'groups', 'knowledge_date_column', 'prefix']:
            if key not in aggregation_config:
                raise ValueError(
                    '{} required as key: aggregation config: {}'
                    .format(key, aggregation_config)
                )

    def _validate_aggregates(self, aggregation_config):
        if 'aggregates' not in aggregation_config \
                and 'categoricals' not in aggregation_config \
                and 'array_categoricals' not in aggregation_config:
            raise ValueError(
                'Need either aggregates, categoricals, or array_categoricals' +
                ' in {}'.format(aggregation_config)
            )

    def _validate_categoricals(self, categoricals):
        for categorical in categoricals:
            if 'choice_query' in categorical:
                logging.info('Validating choice query')

                try:
                    with self.db_engine.begin() as conn:
                        conn.execute('explain {}'.format(categorical['choice_query']))
                except Exception as exc:
                    raise ValueError(
                        'choice query does not run. \n'
                        'choice query: "{}"\n'
                        'Full error: {}'
                        .format(categorical['choice_query'], exc)
                    )

    def _validate_from_obj(self, from_obj):
        logging.info('Validating from_obj')
        try:
            with self.db_engine.begin() as conn:
                conn.execute('explain select * from {}'.format(from_obj))
        except Exception as exc:
            raise ValueError(
                'from_obj query does not run. \n'
                'from_obj: "{}"\n'
                'Full error: {}'
                .format(from_obj, exc)
            )

    def _validate_time_intervals(self, intervals):
        logging.info('Validating time intervals')
        for interval in intervals:
            if interval != 'all':
                convert_str_to_relativedelta(interval)

    def _validate_groups(self, groups):
        if 'entity_id' not in groups:
            raise ValueError('One of the aggregation groups is required to be entity_id')

    def _validate_imputation_rule(self, aggregate_type, impute_rule):
        """Validate the imputation rule for a given aggregation type."""
        # dictionary of imputation type : required parameters
        valid_imputations = {
            'all': {
                'mean': [],
                'constant': ['value'],
                'zero': [],
                'zero_noflag': [],
                'error': []
            },
            'aggregates': {
                'binary_mode': [],
            },
            'categoricals': {
                'null_category': []
            }
        }
        valid_imputations['array_categoricals'] = valid_imputations['categoricals']

        # the valid imputation rules for the specific aggregation type being checked
        valid_types = dict(valid_imputations['all'], **valid_imputations[aggregate_type])

        # no imputation rule was specified
        if 'type' not in impute_rule.keys():
            raise ValueError('Imputation type must be specified')

        # a rule was specified, but not valid for this type of aggregate
        if impute_rule['type'] not in valid_types.keys():
            raise ValueError('Invalid imputation type %s for %s'
                             % (impute_rule['type'], aggregate_type))

        # check that all required parameters exist in the keys of the imputation rule
        required_params = valid_types[impute_rule['type']]
        for param in required_params:
            if param not in impute_rule.keys():
                raise ValueError('Missing param %s for %s' % (param, impute_rule['type']))

    def _validate_imputations(self, aggregation_config):
        """Validate the imputation rules in an aggregation config, looping
        through all three types of aggregates. Most of the work here is
        done by _validate_imputation_rule() to check the requirements of
        each imputation rule found
        """
        agg_types = ['aggregates', 'categoricals', 'array_categoricals']

        for agg_type in agg_types:
            # base_imp are the top-level rules, `such as aggregates_imputation`
            base_imp = aggregation_config.get(agg_type+'_imputation', {})

            # loop through the individual aggregates
            for agg in aggregation_config.get(agg_type, []):
                # combine any aggregate-level imputation rules with top-level ones
                imp_dict = dict(base_imp, **agg.get('imputation', {}))

                # imputation rules are metric-specific, so check each metric's rule
                for metric in agg['metrics']:
                    # metric rules may be defined by the metric name (e.g., 'max')
                    # or with the 'all' catch-all, with named metrics taking
                    # precedence. If we fall back to {}, the rule validator will
                    # error out on no metric found.
                    impute_rule = imp_dict.get(metric, imp_dict.get('all', {}))
                    self._validate_imputation_rule(agg_type, impute_rule)

    def _validate_aggregation(self, aggregation_config):
        logging.info('Validating aggregation config %s', aggregation_config)
        self._validate_keys(aggregation_config)
        self._validate_aggregates(aggregation_config)
        self._validate_categoricals(aggregation_config.get('categoricals', []))
        self._validate_from_obj(aggregation_config['from_obj'])
        self._validate_time_intervals(aggregation_config['intervals'])
        self._validate_groups(aggregation_config['groups'])
        self._validate_imputations(aggregation_config)

    def validate(self, feature_aggregation_config):
        """Validate a feature aggregation config applied to this object

        The validations range from basic type checks, key presence checks,
        as well as validating the sql in from objects.

        Args:
            feature_aggregation_config (list) all values, except for feature
                date, necessary to instantiate a collate.SpacetimeAggregation

        Raises: ValueError if any part of the config is found to be invalid
        """
        for aggregation in feature_aggregation_config:
            self._validate_aggregation(aggregation)

    def _compute_choices(self, choice_query):
        if choice_query not in self.categorical_cache:
            with self.db_engine.begin() as conn:
                self.categorical_cache[choice_query] = [
                    row[0] for row in conn.execute(choice_query)
                ]

            logging.info(
                'Computed list of categoricals: %s for choice query: %s',
                self.categorical_cache[choice_query],
                choice_query
            )

        return self.categorical_cache[choice_query]

    def _build_choices(self, categorical):
        logging.info(
            'Building categorical choices for column %s, metrics %s',
            categorical['column'],
            categorical['metrics']
        )
        if 'choices' in categorical:
            logging.info('Found list of configured choices: %s', categorical['choices'])
            return categorical['choices']
        else:
            return self._compute_choices(categorical['choice_query'])

    def _build_categoricals(self, categorical_config, impute_rules):
        # TODO: only include null flag where necessary
        return [
            Categorical(
                col=categorical['column'],
                choices=self._build_choices(categorical),
                function=categorical['metrics'],
                impute_rules=dict(
                    impute_rules,
                    coltype='categorical',
                    **categorical.get('imputation', {})
                ),
                include_null=True
            )
            for categorical in categorical_config
        ]

    def _build_array_categoricals(self, categorical_config, impute_rules):
        # TODO: only include null flag where necessary
        return [
            Compare(
                col=categorical['column'],
                op='@>',
                choices={
                    choice: "array['{}'::varchar]".format(choice)
                    for choice in
                    self._build_choices(categorical)
                },
                function=categorical['metrics'],
                impute_rules=dict(
                    impute_rules,
                    coltype='array_categorical',
                    **categorical.get('imputation', {})
                ),
                op_in_name=False,
                quote_choices=False,
                include_null=True
            )
            for categorical in categorical_config
        ]

    def _aggregation(self, aggregation_config, feature_dates, state_table):
        logging.info(
            'Building collate.SpacetimeAggregation for config %s and as_of_dates %s',
            aggregation_config,
            feature_dates
        )

        # read top-level imputation rules from the aggregation config; we'll allow
        # these to be overridden by imputation rules at the individual feature
        # level as those get parsed as well
        agimp = aggregation_config.get('aggregates_imputation', {})
        catimp = aggregation_config.get('categoricals_imputation', {})
        arrcatimp = aggregation_config.get('array_categoricals_imputation', {})

        aggregates = [
            Aggregate(
                aggregate['quantity'],
                aggregate['metrics'],
                dict(agimp, coltype='aggregate', **aggregate.get('imputation', {}))
            )
            for aggregate in aggregation_config.get('aggregates', [])
        ]
        logging.info('Found %s quantity aggregates', len(aggregates))
        categoricals = self._build_categoricals(
            aggregation_config.get('categoricals', []),
            catimp
        )
        logging.info('Found %s categorical aggregates', len(categoricals))
        array_categoricals = self._build_array_categoricals(
            aggregation_config.get('array_categoricals', []),
            arrcatimp
        )
        logging.info('Found %s array categorical aggregates', len(array_categoricals))
        return SpacetimeAggregation(
            aggregates + categoricals + array_categoricals,
            from_obj=aggregation_config['from_obj'],
            intervals=aggregation_config['intervals'],
            groups=aggregation_config['groups'],
            dates=feature_dates,
            state_table=state_table,
            state_group=self.entity_id_column,
            date_column=aggregation_config['knowledge_date_column'],
            output_date_column='as_of_date',
            input_min_date=self.feature_start_time,
            schema=self.features_schema_name,
            prefix=aggregation_config['prefix']
        )

    def aggregations(self, feature_aggregation_config, feature_dates, state_table):
        """Creates collate.SpacetimeAggregations from the given arguments

        Args:
            feature_aggregation_config (list) all values, except for feature
                date, necessary to instantiate a collate.SpacetimeAggregation
            feature_dates (list) dates to generate features as of
            state_table (string) schema.table_name for state table with all entity/date pairs

        Returns: (list) collate.SpacetimeAggregations
        """
        return [
            self._aggregation(aggregation_config, feature_dates, state_table)
            for aggregation_config in feature_aggregation_config
        ]

    def generate_all_table_tasks(self, aggregations, task_type):
        """Generates SQL commands for creating, populating, and indexing
        feature group tables

        Args:
            aggregations (list) collate.SpacetimeAggregation objects
            type (str) either 'aggregation' or 'imputation'

        Returns: (dict) keys are group table names, values are themselves dicts,
            each with keys for different stages of table creation (prepare, inserts, finalize)
            and with values being lists of SQL commands
        """

        logging.debug('---------------------')

        # pick the method to use for generating tasks depending on whether we're
        # building the aggregations or imputations
        if task_type == 'aggregation':
            task_generator = self._generate_agg_table_tasks_for
            logging.debug('---------FEATURE GENERATION------------')
        elif task_type == 'imputation':
            task_generator = self._generate_imp_table_tasks_for
            logging.debug('---------FEATURE IMPUTATION------------')
        else:
            raise ValueError('Table task type must be aggregation or imputation')

        logging.debug('---------------------')

        table_tasks = OrderedDict()
        for aggregation in aggregations:
            table_tasks.update(task_generator(aggregation))
        logging.info('Created %s tables', len(table_tasks.keys()))
        return table_tasks

    def create_features_before_imputation(self, feature_aggregation_config, feature_dates):
        """Create features before imputation for a set of dates"""
        all_tasks = self.generate_all_table_tasks(
            self.aggregations(
                feature_aggregation_config,
                feature_dates,
                state_table=None
            ),
            task_type='aggregation'
        )
        logging.info('Generated a total of %s table tasks', len(all_tasks))
        for task_num, task in enumerate(all_tasks.values(), 1):
            prepares = task.get('prepare', [])
            inserts = task.get('inserts', [])
            finalize = task.get('finalize', [])
            logging.info('------------------')
            logging.info('TASK %s ', task_num)
            logging.info('%s prepare queries, %s insert queries, %s finalize queries', len(prepares), len(inserts), len(finalize))
            logging.info('------------------')
            logging.info('')
            logging.info('------------------')
            logging.info('PREPARATION QUERIES')
            logging.info('------------------')
            for query_num, query in enumerate(prepares, 1):
                logging.info('')
                logging.info('prepare query %s: %s', query_num, sqlparse.format(str(query), reindent=True))
            logging.info('------------------')
            logging.info('INSERT QUERIES')
            logging.info('------------------')
            for query_num, query in enumerate(inserts, 1):
                logging.info('')
                logging.info('insert query %s: %s', query_num, sqlparse.format(str(query), reindent=True))
            logging.info('------------------')
            logging.info('FINALIZE QUERIES')
            logging.info('------------------')
            for query_num, query in enumerate(finalize, 1):
                logging.info('')
                logging.info('finalize query %s: %s', query_num, sqlparse.format(str(query), reindent=True))
            self.process_table_task(task)

    def create_all_tables(self, feature_aggregation_config, feature_dates, state_table):
        """Create all feature tables.

        First builds the aggregation tables, and then performs
        imputation on any null values, (requiring a two-step process to
        determine which columns contain nulls after the initial
        aggregation tables are built).

        Args:
            feature_aggregation_config (list) all values, except for
                feature date, necessary to instantiate a
                `collate.SpacetimeAggregation`
            feature_dates (list) dates to generate features as of
            state_table (string) schema.table_name for state table with
                all entity/date pairs

        Returns: (list) table names

        """
        aggs = self.aggregations(
            feature_aggregation_config,
            feature_dates,
            state_table
        )

        # first, generate and run table tasks for aggregations
        table_tasks_aggregate = self.generate_all_table_tasks(
            aggs,
            task_type='aggregation'
        )
        self.process_table_tasks(table_tasks_aggregate)

        # second, perform the imputations (this will query the tables
        # constructed above to identify features containing nulls)
        table_tasks_impute = self.generate_all_table_tasks(
            aggs,
            task_type='imputation'
        )
        impute_keys = self.process_table_tasks(table_tasks_impute)

        # double-check that the imputation worked and no nulls remain
        # in the data:
        nullcols = []
        with self.db_engine.begin() as conn:
            for agg in aggs:
                results = conn.execute(agg.find_nulls(imputed=True))
                null_counts = results.first().items()
                nullcols += [col for (col, val) in null_counts if val > 0]

        if len(nullcols) > 0:
            raise ValueError(
                "Imputation failed for {} columns. Null values remain in: {}"
                .format(len(nullcols), nullcols)
            )

        return impute_keys

    def process_table_task(self, task):
        self.run_commands(task.get('prepare', []))
        self.run_commands(task.get('inserts', []))
        self.run_commands(task.get('finalize', []))

    def process_table_tasks(self, table_tasks):
        for table_name, task in table_tasks.items():
            logging.info('Running feature table queries for %s', table_name)
            self.process_table_task(task)
        return table_tasks.keys()

    def _explain_selects(self, aggregations):
        with self.db_engine.begin() as conn:
            for aggregation in aggregations:
                for selectlist in aggregation.get_selects().values():
                    for select in selectlist:
                        query = 'explain ' + str(select)
                        results = list(conn.execute(query))
                        logging.debug(str(select))
                        logging.debug(results)

    def _clean_table_name(self, table_name):
        # remove the schema and quotes from the name
        return table_name.split('.')[1].replace('"', "")

    def _table_exists(self, table_name):
        try:
            with self.db_engine.begin() as conn:
                conn.execute('select 1 from {}.{} limit 1'
                             .format(self.features_schema_name,
                                     table_name)
                             ).first()
        except sqlalchemy.exc.ProgrammingError:
            return False
        else:
            return True

    def run_commands(self, command_list):
        with self.db_engine.begin() as conn:
            for command in command_list:
                logging.debug('Executing feature generation query: %s', command)
                conn.execute(command)

    def _aggregation_index_query(self, aggregation, imputed=False):
        return 'CREATE INDEX ON {} ({}, {})'.format(
            aggregation.get_table_name(imputed=imputed),
            self.entity_id_column,
            aggregation.output_date_column
        )

    def _aggregation_index_columns(self, aggregation):
        return sorted([group for group in aggregation.groups.keys()] +
                      [aggregation.output_date_column])

    def index_column_lookup(self, aggregations, imputed=True):
        return dict((
            self._clean_table_name(aggregation.get_table_name(imputed=imputed)),
            self._aggregation_index_columns(aggregation)
        ) for aggregation in aggregations)

    def _generate_agg_table_tasks_for(self, aggregation):
        """Generates SQL commands for preparing, populating, and finalizing
        each feature group table in the given aggregation

        Args:
            aggregation (collate.SpacetimeAggregation)

        Returns: (dict) of structure {
            'prepare': list of commands to prepare table for population
            'inserts': list of commands to populate table
            'finalize': list of commands to finalize table after population
        }
        """
        create_schema = aggregation.get_create_schema()
        creates = aggregation.get_creates()
        drops = aggregation.get_drops()
        indexes = aggregation.get_indexes()
        inserts = aggregation.get_inserts()

        if create_schema is not None:
            with self.db_engine.begin() as conn:
                conn.execute(create_schema)

        table_tasks = OrderedDict()
        for group in aggregation.groups:
            group_table = self._clean_table_name(
                aggregation.get_table_name(group=group)
            )
            imputed_table = self._clean_table_name(
                aggregation.get_table_name(imputed=True)
            )
            if self.replace or (
                not self._table_exists(group_table)
                and
                not self._table_exists(imputed_table)
            ):
                table_tasks[group_table] = {
                    'prepare': [drops[group], creates[group]],
                    'inserts': inserts[group],
                    'finalize': [indexes[group]],
                }
                logging.info('Created table tasks for %s', group_table)
            else:
                logging.info(
                    'Skipping feature table creation for %s',
                    group_table
                )
                table_tasks[group_table] = {}
        logging.info('Created table tasks for aggregation')
        if self.replace or (
            not self._table_exists(
                self._clean_table_name(aggregation.get_table_name())
            )
            and
            not self._table_exists(
                self._clean_table_name(aggregation.get_table_name(imputed=True))
            )
        ):
            table_tasks[self._clean_table_name(aggregation.get_table_name())] = {
                'prepare': [aggregation.get_drop(), aggregation.get_create()],
                'inserts': [],
                'finalize': [self._aggregation_index_query(aggregation)],
            }
        else:
            table_tasks[self._clean_table_name(aggregation.get_table_name())] = {}

        return table_tasks

    def _generate_imp_table_tasks_for(self, aggregation, drop_preagg=True):
        """Generate SQL statements for preparing, populating, and
        finalizing imputations, for each feature group table in the
        given aggregation.

        Requires the existance of the underlying feature and aggregation
        tables defined in `_generate_agg_table_tasks_for()`.

        Args:
            aggregation (collate.SpacetimeAggregation)
            drop_preagg: boolean to specify dropping pre-imputation
                tables

        Returns: (dict) of structure {
                'prepare': list of commands to prepare table for population
                'inserts': list of commands to populate table
                'finalize': list of commands to finalize table after population
            }

        """
        table_tasks = OrderedDict()
        imp_tbl_name = self._clean_table_name(
            aggregation.get_table_name(imputed=True)
        )

        if not self.replace and self._table_exists(imp_tbl_name):
            logging.info('Skipping imputation table creation for %s',
                         imp_tbl_name)
            table_tasks[imp_tbl_name] = {}
            return table_tasks

        if not aggregation.state_table:
            logging.warning(
                'No state table available to aggregation, cannot create imputation table for %s',
                imp_tbl_name
            )
            table_tasks[imp_tbl_name] = {}
            return table_tasks

        # excute query to find columns with null values and create lists of columns
        # that do and do not need imputation when creating the imputation table
        with self.db_engine.begin() as conn:
            results = conn.execute(aggregation.find_nulls())
            null_counts = results.first().items()
        impute_cols = [col for (col, val) in null_counts if val > 0]
        nonimpute_cols = [col for (col, val) in null_counts if val == 0]

        # table tasks for imputed aggregation table, most of the work is done here
        # by collate's get_impute_create()
        table_tasks[imp_tbl_name] = {
            'prepare': [
                aggregation.get_drop(imputed=True),
                aggregation.get_impute_create(
                    impute_cols=impute_cols,
                    nonimpute_cols=nonimpute_cols
                )
            ],
            'inserts': [],
            'finalize': [self._aggregation_index_query(aggregation, imputed=True)]
        }
        logging.info('Created table tasks for imputation: %s', imp_tbl_name)

        # do some cleanup:
        # drop the group-level and aggregation tables, just leaving the
        # imputation table if drop_preagg=True
        if drop_preagg:
            drops = aggregation.get_drops()
            table_tasks[imp_tbl_name]['finalize'] += (list(drops.values()) +
                                                      [aggregation.get_drop()])
            logging.info('Added drop table cleanup tasks: %s', imp_tbl_name)

        return table_tasks



# -*- coding: utf-8 -*-
from itertools import chain
import sqlalchemy.sql.expression as ex

from .sql import make_sql_clause
from .collate import Aggregation


class SpacetimeAggregation(Aggregation):
    def __init__(self, aggregates, groups, intervals, from_obj, dates,
                 state_table, state_group=None,
                 prefix=None, suffix=None, schema=None, date_column=None,
                 output_date_column=None, input_min_date=None):
        """
        Args:
            intervals: the intervals to aggregate over. either a list of
                datetime intervals, e.g. ["1 month", "1 year"], or
                a dictionary of group : intervals pairs where
                group is a group in groups and intervals is a collection
                of datetime intervals, e.g. {"address_id": ["1 month", "1 year]}
            dates: list of PostgreSQL date strings,
                e.g. ["2012-01-01", "2013-01-01"]
            state_table: schema.table to query for valid state_group/date combinations
            state_group: the group level found in the state table (e.g., "entity_id")
            date_column: name of date column in from_obj, defaults to "date"
            output_date_column: name of date column in aggregated output, defaults to "date"
            input_min_date: minimum date for which rows shall be included, defaults
                to no absolute time restrictions on the minimum date of included rows

        For all other arguments see collate.Aggregation
        """
        Aggregation.__init__(self,
                             aggregates=aggregates,
                             from_obj=from_obj,
                             groups=groups,
                             state_table=state_table,
                             state_group=state_group,
                             prefix=prefix,
                             suffix=suffix,
                             schema=schema)

        if isinstance(intervals, dict):
            self.intervals = intervals
        else:
            self.intervals = {g: intervals for g in self.groups}
        self.dates = dates
        self.date_column = date_column if date_column else "date"
        self.output_date_column = output_date_column if output_date_column else "date"
        self.input_min_date = input_min_date

    def _state_table_sub(self):
        """Helper function to ensure we only include state table records
        in our set of input dates and after the input_min_date.
        """
        datestr = ', '.join(["'%s'::date" % dt for dt in self.dates])
        mindtstr = " AND %s >= '%s'::date" %\
            (self.output_date_column, self.input_min_date)\
            if self.input_min_date is not None else ""
        return """(
        SELECT *
        FROM {st}
        WHERE {datecol} IN ({datestr})
        {mindtstr})""".format(
            st=self.state_table,
            datecol=self.output_date_column,
            datestr=datestr,
            mindtstr=mindtstr
        )

    def _get_aggregates_sql(self, interval, date, group):
        """
        Helper for getting aggregates sql
        Args:
            interval: SQL time interval string, or "all"
            date: SQL date string
            group: group clause, for naming columns
        Returns: collection of aggregate column SQL strings
        """
        if interval != 'all':
            when = "{date_column} >= '{date}'::date - interval '{interval}'".format(
                    interval=interval, date=date, date_column=self.date_column)
        else:
            when = None

        prefix = "{prefix}_{group}_{interval}_".format(
                prefix=self.prefix, interval=interval,
                group=group)

        return chain(*[a.get_columns(when, prefix, format_kwargs={"collate_date": date,
                                                                  "collate_interval": interval})
                       for a in self.aggregates])

    def get_selects(self):
        """
        Constructs select queries for this aggregation

        Returns: a dictionary of group : queries pairs where
            group are the same keys as groups
            queries is a list of Select queries, one for each date in dates
        """
        queries = {}

        for group, groupby in self.groups.items():
            intervals = self.intervals[group]
            queries[group] = []
            for date in self.dates:
                columns = [groupby,
                           ex.literal_column("'%s'::date"
                                             % date).label(self.output_date_column)]
                columns += list(chain(*[self._get_aggregates_sql(
                        i, date, group) for i in intervals]))

                gb_clause = make_sql_clause(groupby, ex.literal_column)
                query = ex.select(columns=columns, from_obj=self.from_obj)\
                          .group_by(gb_clause)
                query = query.where(self.where(date, intervals))

                queries[group].append(query)

        return queries

    def get_imputation_rules(self):
        """
        Constructs a dictionary to lookup an imputation rule from an associated
        column name.

        Returns: a dictionary of column : imputation_rule pairs
        """
        imprules = {}
        for group, groupby in self.groups.items():
            for interval in self.intervals[group]:
                prefix = "{prefix}_{group}_{interval}_".format(
                        prefix=self.prefix, interval=interval,
                        group=group)
                for a in self.aggregates:
                    imprules.update(a.column_imputation_lookup(prefix=prefix))
        return imprules

    def where(self, date, intervals):
        """
        Generates a WHERE clause
        Args:
            date: the end date
            intervals: intervals

        Returns: a clause for filtering the from_obj to be between the date and
            the greatest interval
        """
        # upper bound
        w = "{date_column} < '{date}'".format(
                            date_column=self.date_column, date=date)

        # lower bound (if possible)
        if 'all' not in intervals:
            greatest = "greatest(%s)" % str.join(
                    ",", ["interval '%s'" % i for i in intervals])
            min_date = "'{date}'::date - {greatest}".format(date=date, greatest=greatest)
            w += "AND {date_column} >= {min_date}".format(
                    date_column=self.date_column, min_date=min_date)
        if self.input_min_date is not None:
            w += "AND {date_column} >= '{bot}'::date".format(
                    date_column=self.date_column, bot=self.input_min_date)
        return ex.text(w)

    def get_indexes(self):
        """
        Generate create index queries for this aggregation

        Returns: a dictionary of group : index pairs where
            group are the same keys as groups
            index is a raw create index query for the corresponding table
        """
        return {group: "CREATE INDEX ON %s (%s, %s);" %
                (self.get_table_name(group), groupby, self.output_date_column)
                for group, groupby in self.groups.items()}

    def get_join_table(self):
        """
        Generates a join table, consisting of an entry for each combination of
        groups and dates in the from_obj
        """
        groups = list(self.groups.values())
        intervals = list(set(chain(*self.intervals.values())))

        queries = []
        for date in self.dates:
            columns = groups + [ex.literal_column("'%s'::date" % date).label(
                    self.output_date_column)]
            queries.append(ex.select(columns, from_obj=self.from_obj)
                             .where(self.where(date, intervals))
                             .group_by(*groups))

        return str.join("\nUNION ALL\n", map(str, queries))

    def get_create(self, join_table=None):
        """
        Generate a single aggregation table creation query by joining
            together the results of get_creates()
        Returns: a CREATE TABLE AS query
        """
        if not join_table:
            join_table = '(%s) t1' % self.get_join_table()
        query = "SELECT * FROM %s\n" % join_table
        for group, groupby in self.groups.items():
            query += " LEFT JOIN %s USING (%s, %s)" % (
                    self.get_table_name(group), groupby, self.output_date_column)

        return "CREATE TABLE %s AS (%s);" % (self.get_table_name(), query)

    def validate(self, conn):
        """
        SpacetimeAggregations ensure that no intervals extend beyond the absolute
        minimum time.
        """
        if self.input_min_date is not None:
            all_intervals = set(*self.intervals.values())
            for date in self.dates:
                for interval in all_intervals:
                    if interval == "all":
                        continue
                    # This could be done more efficiently all at once, but doing
                    # it this way allows for nicer error messages.
                    r = conn.execute("select ('%s'::date - '%s'::interval) < '%s'::date" %
                                     (date, interval, self.input_min_date))
                    if r.fetchone()[0]:
                        raise ValueError(
                            "date '%s' - '%s' is before input_min_date ('%s')" %
                            (date, interval, self.input_min_date))
                    r.close()
        for date in self.dates:
            r = conn.execute("select count(*) from %s where %s = '%s'::date" %
                             (self.state_table, self.output_date_column, date))
            if r.fetchone()[0] == 0:
                raise ValueError(
                    "date '%s' is not present in states table ('%s')" %
                    (date, self.state_table))
            r.close()

    def find_nulls(self, imputed=False):
        """
        Generate query to count number of nulls in each column in the aggregation table

        Returns: a SQL SELECT statement
        """
        query_template = """
            SELECT {cols}
            FROM {state_tbl} t1
            LEFT JOIN {aggs_tbl} t2 USING({group}, {date_col})
            """
        cols_sql = ',\n'.join([
            """SUM(CASE WHEN "{col}" IS NULL THEN 1 ELSE 0 END) AS "{col}" """.format(col=column)
            for column in self.get_imputation_rules().keys()
            ])

        return query_template.format(
                cols=cols_sql,
                state_tbl=self._state_table_sub(),
                aggs_tbl=self.get_table_name(imputed=imputed),
                group=self.state_group,
                date_col=self.output_date_column
            )

    def get_impute_create(self, impute_cols, nonimpute_cols):
        """
        Generates the CREATE TABLE query for the aggregation table with imputation.

        Args:
            impute_cols: a list of column names with null values
            nonimpute_cols: a list of column names without null values

        Returns: a CREATE TABLE AS query
        """

        # key columns and date column
        query = "SELECT %s, %s" % (
            ', '.join(map(str, self.groups.values())),
            self.output_date_column
            )

        # columns with imputation filling as needed
        query += self._get_impute_select(
            impute_cols,
            nonimpute_cols,
            partitionby=self.output_date_column
        )

        # imputation starts from the state table and left joins into the aggregation table
        query += "\nFROM %s t1" % self._state_table_sub()
        query += "\nLEFT JOIN %s t2 USING(%s, %s)" % (
            self.get_table_name(),
            self.state_group,
            self.output_date_column
            )

        return "CREATE TABLE %s AS (%s)" % (self.get_table_name(imputed=True), query)



"""Changing column names and removing redundancies in table names

Revision ID: 2446a931de7a
Revises: 89a8ce240bae
Create Date: 2018-05-24 17:07:20.567789

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '2446a931de7a'
down_revision = '89a8ce240bae'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('ALTER TABLE test_results.test_evaluations RENAME TO evaluations;' +
               'ALTER TABLE test_results.test_predictions RENAME TO predictions;' +
               'ALTER TABLE train_results.train_evaluations RENAME TO evaluations;' +
               'ALTER TABLE train_results.train_predictions RENAME TO predictions;'
               )

    op.alter_column('matrices', 'n_examples', new_column_name='num_observations', schema='model_metadata')
    op.alter_column('model_groups', 'model_parameters', new_column_name='hyperparameters', schema='model_metadata')
    op.alter_column('models', 'model_parameters', new_column_name='hyperparameters', schema='model_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('ALTER TABLE test_results.evaluations RENAME TO test_evaluations;' +
               'ALTER TABLE test_results.predictions RENAME TO test_predictions;' +
               'ALTER TABLE train_results.evaluations RENAME TO train_evaluations;' +
               'ALTER TABLE train_results.predictions RENAME TO train_predictions;'
               )

    op.alter_column('matrices', 'num_observations', new_column_name='n_examples', schema='model_metadata')
    op.alter_column('model_groups', 'hyperparameters', new_column_name='model_parameters', schema='model_metadata')
    op.alter_column('models', 'hyperparameters', new_column_name='model_parameters', schema='model_metadata')

    # ### end Alembic commands ###



class BaseRules(object):
    def __init__(self):
        self._metric = None
        self._parameter = None
        self.shared_parameters = []
        self.selection_rules = []

    def _does_parameters_exist(self, params_dict):
        return params_dict in self.shared_parameters

    def _does_selection_rule_exisit(self, rule_dict):
        return rule_dict in self.selection_rules

    def _append(self, params_dict, rule_dict):
        if not self._does_parameters_exist(params_dict):
            self.shared_parameters.append(params_dict)
        if not self._does_selection_rule_exisit(rule_dict):
            self.selection_rules.append(rule_dict)

    def create(self):
        return [{
            'shared_parameters': self.shared_parameters,
            'selection_rules': self.selection_rules
        }]


class SimpleRuleMaker(BaseRules):
    def add_rule_best_current_value(self, metric=None, parameter=None, n=1):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {'metric': self._metric, 'parameter': self._parameter}
        rule_dict = {'name': 'best_current_value', 'n': n}
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_best_average_value(self, metric=None, parameter=None, n=1):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {'metric': self._metric, 'parameter': self._parameter}
        rule_dict = {'name': 'best_average_value', 'n': n}
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_lowest_metric_variance(self, metric=None, parameter=None, n=1):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {'metric': self._metric, 'parameter': self._parameter}
        rule_dict = {'name': 'lowest_metric_variance', 'n': n}
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_most_frequent_best_dist(self, metric=None, parameter=None, n=1,
                                         dist_from_best_case=[0.01, 0.05, 0.1, 0.15]):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {'metric': self._metric, 'parameter': self._parameter}
        rule_dict = {
            'name': 'most_frequent_best_dist',
            'dist_from_best_case': dist_from_best_case,
            'n': n
        }
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_best_avg_recency_weight(self, metric=None, parameter=None, n=1,
                                         curr_weight=[1.5, 2.0, 5.0], decay_type=['linear']):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {'metric': metric, 'parameter': parameter}
        rule_dict = {
            'name': 'best_avg_recency_weight',
            'curr_weight': curr_weight,
            'decay_type': decay_type,
            'n': n
        }
        self._append(params_dict, rule_dict)
        return self.create()

    def add_rule_best_avg_var_penalized(self, metric=None, parameter=None, stdev_penalty=0.5, n=1):
        if metric is not None:
            self._metric = metric
        if parameter is not None:
            self._parameter = parameter
        params_dict = {'metric': metric, 'parameter': parameter}
        rule_dict = {
            'name': 'best_avg_var_penalized',
            'stdev_penalty': stdev_penalty,
            'n': n
        }
        self._append(params_dict, rule_dict)
        return self.create()


class RandomGroupRuleMaker(BaseRules):
    def __init__(self, n=1):
        self.shared_parameters = [{}]
        self.selection_rules = [{'name': 'random_model_group', 'n': n}]


class TwoMetricsRuleMaker(BaseRules):
    def add_rule_best_average_two_metrics(self, metric1='precision@', parameter1='100_abs',
                                          metric2='recall@', parameter2='300_abs',
                                          metric1_weight=[0.5], n=1):
        params_dict = {
            'metric1': metric1,
            'parameter1': parameter1
        }
        rule_dict = {
            'name': 'best_average_two_metrics',
            'metric1_weight': metric1_weight,
            'metric2': [metric2],
            'parameter2': [parameter2],
            'n': n
        }
        self._append(params_dict, rule_dict)


def create_selection_grid(*args):
    return list(map(lambda r: r.create()[0], args))



import logging


class FeatureGroup(dict):
    def __init__(self, *args, **kwargs):
        try:
            self._names = [kwargs.pop('name')]
        except KeyError:
            self._names = []
        try:
            features = kwargs.pop('features_by_table')
        except KeyError:
            features = {}
        super(FeatureGroup, self).__init__(*args, **features)

    @property
    def names(self):
        return self._names

    def update(self, other_group):
        super(FeatureGroup, self).update(other_group)
        self._names += other_group.names


def table_subsetter(config_item, table, features):
    "Return features matching a given table"
    if table == config_item:
        return features
    else:
        return []


def prefix_subsetter(config_item, table, features):
    "Return features matching a given prefix"
    return [
        feature
        for feature in features
        if feature.startswith(config_item)
    ]


def all_subsetter(config_item, table, features):
    return features


class FeatureGroupCreator(object):
    """Divides a feature dictionary into groups based on given criteria"""
    subsetters = {
        'tables': table_subsetter,
        'prefix': prefix_subsetter,
        'all': all_subsetter,
    }

    def __init__(self, definition):
        """
        Args:
            definition (dict) rules for generating feature groups
                Each key must correspond to a key in self.subsetters
                Each value (a list) must be understood by the subsetter
        """
        self.definition = definition

    def validate(self):
        """Validate the object's configuration

        """
        if not isinstance(self.definition, dict):
            raise ValueError('Feature Group Definition must be a dictionary')

        for subsetter_name, value in self.definition.items():
            if subsetter_name not in self.subsetters:
                raise ValueError('Unknown subsetter %s received', subsetter_name)
            if not hasattr(value, '__iter__') or isinstance(value, (str, bytes)):
                raise ValueError('Each value in FeatureGroupCreator must be '
                                 'iterable and not a string')

    def subsets(self, feature_dictionary):
        """Generate subsets of a feature dict

        Args:
            feature_dictionary (dict) tables and the features contained in each

            The feature dictionary is meant to be keyed on source table. Example:

            {
                'feature_table_one': ['feature_one', feature_two'],
                'feature_table_two': ['feature_three', 'feature_four'],
            }

        Returns: (list) subsets of the feature dictionary, in the same
            table-based structure
        """
        logging.info(
            'Creating feature groups. config: %s, Master feature dictionary: %s',
            self.definition,
            feature_dictionary
        )
        subsets = []
        for name, config in sorted(self.definition.items()):
            for config_item in config:
                subset = FeatureGroup(name='{}: {}'.format(name, config_item))
                for table, features in feature_dictionary.items():
                    matching_features =\
                        self.subsetters[name](config_item, table, features)
                    logging.info(
                        'Matching features for config item %s, table %s: %s',
                        config_item,
                        table,
                        matching_features
                    )
                    if len(matching_features) > 0:
                        subset[table] = matching_features
                    else:
                        logging.warning(
                            'No matching features found for config item %s, '
                            'table %s, master features %s',
                            config_item,
                            table,
                            features
                        )

                subsets.append(subset)
        if not any(subset for subset in subsets if any(subset)):
            logging.warning('No matching feature groups available.')
        logging.info('Found %s total feature subsets', len(subsets))
        return subsets



from triage.experiments import ExperimentBase


class SingleThreadedExperiment(ExperimentBase):
    def process_query_tasks(self, query_tasks):
        self.feature_generator.process_table_tasks(query_tasks)

    def process_matrix_build_tasks(self, matrix_build_tasks):
        self.matrix_builder.build_all_matrices(matrix_build_tasks)

    def process_train_tasks(self, train_tasks):
        return [self.trainer.process_train_task(**train_task) for train_task in train_tasks]

    def process_model_test_tasks(self, test_tasks):
        return [self.tester.process_model_test_task(**test_task) for test_task in test_tasks]



import csv
import datetime
import hashlib
import json
import random
import tempfile

import postgres_copy
import sqlalchemy
from retrying import retry
from sqlalchemy.orm import sessionmaker

from triage.component.results_schema import Experiment, Model


def filename_friendly_hash(inputs):
    def dt_handler(x):
        if isinstance(x, datetime.datetime) or isinstance(x, datetime.date):
            return x.isoformat()
        raise TypeError("Unknown type")
    return hashlib.md5(
        json.dumps(inputs, default=dt_handler, sort_keys=True)
            .encode('utf-8')
    ).hexdigest()


def retry_if_db_error(exception):
    return isinstance(exception, sqlalchemy.exc.OperationalError)


DEFAULT_RETRY_KWARGS = {
    'retry_on_exception': retry_if_db_error,
    'wait_exponential_multiplier': 1000,  # wait 2^x*1000ms between each retry
    'stop_max_attempt_number': 14,
    # with this configuration, last wait will be ~2 hours
    # for a total of ~4.5 hours waiting
}


db_retry = retry(**DEFAULT_RETRY_KWARGS)


@db_retry
def save_experiment_and_get_hash(config, db_engine):
    experiment_hash = filename_friendly_hash(config)
    session = sessionmaker(bind=db_engine)()
    session.merge(Experiment(
        experiment_hash=experiment_hash,
        config=config
    ))
    session.commit()
    session.close()
    return experiment_hash


class Batch:
    # modified from
    # http://codereview.stackexchange.com/questions/118883/split-up-an-iterable-into-batches
    def __init__(self, iterable, limit=None):
        self.iterator = iter(iterable)
        self.limit = limit
        try:
            self.current = next(self.iterator)
        except StopIteration:
            self.on_going = False
        else:
            self.on_going = True

    def group(self):
        yield self.current
        # start enumerate at 1 because we already yielded the last saved item
        for num, item in enumerate(self.iterator, 1):
            self.current = item
            if num == self.limit:
                break
            yield item
        else:
            self.on_going = False

    def __iter__(self):
        while self.on_going:
            yield self.group()


def sort_predictions_and_labels(predictions_proba, labels, sort_seed):
    random.seed(sort_seed)
    predictions_proba_sorted, labels_sorted = zip(*sorted(
        zip(predictions_proba, labels),
        key=lambda pair: (pair[0], random.random()), reverse=True)
    )
    return predictions_proba_sorted, labels_sorted


@db_retry
def retrieve_model_id_from_hash(db_engine, model_hash):
    """Retrieves a model id from the database that matches the given hash

    Args:
        db_engine (sqlalchemy.engine) A database engine
        model_hash (str) The model hash to lookup

    Returns: (int) The model id (if found in DB), None (if not)
    """
    session = sessionmaker(bind=db_engine)()
    try:
        saved = session.query(Model)\
            .filter_by(model_hash=model_hash)\
            .one_or_none()
        return saved.model_id if saved else None
    finally:
        session.close()


@db_retry
def save_db_objects(db_engine, db_objects):
    """Saves a collection of SQLAlchemy model objects to the database using a COPY command

    Args:
        db_engine (sqlalchemy.engine)
        db_objects (list) SQLAlchemy model objects, corresponding to a valid table
    """
    with tempfile.TemporaryFile(mode='w+') as f:
        writer = csv.writer(f, quoting=csv.QUOTE_MINIMAL)
        for db_object in db_objects:
            writer.writerow([
                getattr(db_object, col.name)
                for col in db_object.__table__.columns
            ])
        f.seek(0)
        postgres_copy.copy_from(f, type(db_objects[0]), db_engine, format='csv')



from six import string_types


def convert_to_list(x):
    """Given an object, if it is not a list, convert it to a list.

    :param x: an object to be converted to a list
    :type x: object

    :return: x as a list
    :rtype: list

    """
    if isinstance(x, string_types):
        return [x]

    try:
        iter(x)
    except TypeError:
        return [x]
    else:
        return list(x)



# coding: utf-8

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.linear_model import LogisticRegression

from .transformers import CutOff


class ScaledLogisticRegression(BaseEstimator, ClassifierMixin):
    """
    An in-place replacement for the scikit-learn's LogisticRegression.

    It incorporates the MaxMinScaler, and the CutOff as preparations
    for the  logistic regression.
    """
    def __init__(self, penalty='l2', dual=False, tol=1e-4, C=1.0,
                 fit_intercept=True, intercept_scaling=1, class_weight=None,
                 random_state=None, solver='liblinear', max_iter=100,
                 multi_class='ovr', verbose=0, warm_start=False, n_jobs=1):
        self.penalty = penalty
        self.dual = dual
        self.tol = tol
        self.C = C
        self.fit_intercept = fit_intercept
        self.intercept_scaling = intercept_scaling
        self.class_weight = class_weight
        self.random_state = random_state
        self.solver = solver
        self.max_iter = max_iter
        self.multi_class = multi_class
        self.verbose = verbose
        self.warm_start = warm_start
        self.n_jobs = n_jobs

        self.minmax_scaler = MinMaxScaler()
        self.dsapp_cutoff = CutOff()
        self.lr = LogisticRegression(penalty=penalty,
                                     dual=dual,
                                     tol=tol,
                                     C=C,
                                     fit_intercept=fit_intercept,
                                     intercept_scaling=intercept_scaling,
                                     class_weight=class_weight,
                                     random_state=random_state,
                                     solver=solver,
                                     max_iter=max_iter,
                                     multi_class=multi_class,
                                     verbose=verbose,
                                     warm_start=warm_start,
                                     n_jobs=n_jobs)

        self.pipeline = Pipeline([
            ('minmax_scaler', self.minmax_scaler),
            ('dsapp_cutoff', self.dsapp_cutoff),
            ('lr', self.lr)
        ])

    def fit(self, X, y=None):
        self.pipeline.fit(X, y)

        self.min_ = self.pipeline.named_steps['minmax_scaler'].min_
        self.scale_ = self.pipeline.named_steps['minmax_scaler'].scale_
        self.data_min_ = self.pipeline.named_steps['minmax_scaler'].data_min_
        self.data_max_ = self.pipeline.named_steps['minmax_scaler'].data_max_
        self.data_range_ = self.pipeline.named_steps['minmax_scaler'].data_range_

        self.coef_ = self.pipeline.named_steps['lr'].coef_
        self.intercept_ = self.pipeline.named_steps['lr'].intercept_

        self.classes_ = self.pipeline.named_steps['lr'].classes_

        return self

    def predict_proba(self, X):
        return self.pipeline.predict_proba(X)

    def predict_log_proba(self, X):
        return self.pipeline.predict_log_proba(X)

    def predict(self, X):
        return self.pipeline.predict(X)

    def score(self, X, y):
        return self.pipeline.score(X, y)



# -*- coding: utf-8 -*-
"""test_imputation_output

Unit tests for imputation output.

For all available imputation methods, make sure they correctly handle a variety
of cases, including completely null columns and columns entirely null for a given
date, generating the right number of records with no nulls in the output set and
imputed flags as necessary.

"""
import pandas as pd
import pytest
import sqlalchemy
import testing.postgresql

from triage.component.collate import (
    Aggregate,
    available_imputations,
    SpacetimeAggregation,
)

# some imputations require arguments, so specify default values here
# everything in collate.collate.available_imputations needs a record here!
imputation_values = {
    'mean': {
        'aggregate': {'avail': True, 'kwargs': {}},
        'categorical': {'avail': True, 'kwargs': {}}
    },
    'constant': {
        'aggregate': {'avail': True, 'kwargs': {'value': 7}},
        'categorical': {'avail': True, 'kwargs': {'value': 'foo'}}
    },
    'zero': {
        'aggregate': {'avail': True, 'kwargs': {}},
        'categorical': {'avail': True, 'kwargs': {}}
    },
    'zero_noflag': {
        'aggregate': {'avail': True, 'kwargs': {}},
        'categorical': {'avail': True, 'kwargs': {}}
    },
    'null_category': {
        'aggregate': {'avail': False},
        'categorical': {'avail': True, 'kwargs': {}}
    },
    'binary_mode': {
        'aggregate': {'avail': True, 'kwargs': {}},
        'categorical': {'avail': False}
    },
}

states_table = [
    # entity_id, as_of_date
    [1, '2016-01-01'],
    [2, '2016-01-01'],
    [3, '2016-01-01'],
    [4, '2016-01-01'],
    [1, '2016-02-03'],
    [2, '2016-02-03'],
    [3, '2016-02-03'],
    [4, '2016-02-03'],
    [1, '2016-03-14'],
    [2, '2016-03-14'],
    [3, '2016-03-14'],
    [4, '2016-03-14']
]

aggs_table = [
    # entity_id, as_of_date, f1, f2, f3, f4
    [1, '2016-01-01', None, 1, 3, 0],
    [2, '2016-01-01', None, 5, None, 0],
    [3, '2016-01-01', None, 7, 3, 0],
    [4, '2016-01-01', None, 4, 3, 0],
    [1, '2016-02-03', None, None, 3, 0],
    [2, '2016-02-03', None, None, 5, 8],
    [3, '2016-02-03', None, None, None, 2],
    [4, '2016-02-03', None, None, 4, 6],
    [1, '2016-03-14', None, 1, 3, 2],
    [4, '2016-03-14', None, 3, 9, 1]
]

aggs_table_noimp = [
    # entity_id, as_of_date, f5, f6
    [1, '2016-01-01', 3, 0],
    [2, '2016-01-01', None, 0],
    [3, '2016-01-01', 3, 0],
    [4, '2016-01-01', 3, 0],
    [1, '2016-02-03', 3, 0],
    [2, '2016-02-03', 5, 8],
    [3, '2016-02-03', None, 2],
    [4, '2016-02-03', 4, 6],
    [1, '2016-03-14', 3, 2],
    [2, '2016-03-14', None, 2],
    [3, '2016-03-14', None, 1],
    [4, '2016-03-14', 9, 1]
]


def test_available_imputations_coverage():
    assert set(available_imputations.keys()) == \
           set(list(imputation_values.keys()) + ['error'])


@pytest.mark.parametrize(
    ('feat_list', 'exp_imp_cols', 'feat_table'),
    [
        (['f1', 'f2', 'f3', 'f4'], ['f1', 'f2', 'f3', 'f4'], aggs_table),
        (['f5', 'f6'], ['f5'], aggs_table_noimp),
    ]
)
def test_imputation_output(feat_list, exp_imp_cols, feat_table):
    with testing.postgresql.Postgresql() as psql:
        engine = sqlalchemy.create_engine(psql.url())

        engine.execute(
            'create table states (entity_id int, as_of_date date)'
        )
        for state in states_table:
            engine.execute(
                'insert into states values (%s, %s)',
                state
            )

        feat_sql = '\n'.join([', prefix_entity_id_1y_%s_max int' % f for f in feat_list])
        engine.execute(
            '''create table prefix_aggregation (
                entity_id int
                , as_of_date date
                %s
                )''' % feat_sql
        )
        ins_sql = 'insert into prefix_aggregation values (%s, %s'+\
            (', %s' * len(feat_list))+')'
        for rec in feat_table:
            engine.execute(
                ins_sql,
                rec
            )

        for imp in available_imputations.keys():
            # skip error imputation
            if imp == 'error':
                continue

            for coltype in ['aggregate', 'categorical']:
                # only consider
                if not imputation_values[imp][coltype]['avail']:
                    continue

                impargs = imputation_values[imp][coltype]['kwargs']
                aggs = [
                    Aggregate(
                        feat, ['max'], {'coltype': coltype, 'all': dict(type=imp, **impargs)}
                    ) for feat in feat_list
                ]
                st = SpacetimeAggregation(
                        aggregates = aggs,
                        from_obj = 'prefix_events',
                        prefix='prefix',
                        groups = ['entity_id'],
                        intervals = ['1y'],
                        dates = ['2016-01-01', '2016-02-03', '2016-03-14'],
                        state_table = 'states',
                        state_group = 'entity_id',
                        date_column = 'as_of_date',
                        input_min_date = '2000-01-01',
                        output_date_column = 'as_of_date'
                    )

                conn = engine.connect()

                trans = conn.begin()

                # excute query to find columns with null values and create lists of columns
                # that do and do not need imputation when creating the imputation table
                res = conn.execute(st.find_nulls())
                null_counts = list(zip(res.keys(), res.fetchone()))
                impute_cols = [col for col, val in null_counts if val > 0]
                nonimpute_cols = [col for col, val in null_counts if val == 0]

                # sql to drop and create the imputation table
                drop_imp = st.get_drop(imputed=True)
                create_imp = st.get_impute_create(
                            impute_cols=impute_cols,
                            nonimpute_cols=nonimpute_cols
                            )

                # create the imputation table
                conn.execute(drop_imp)
                conn.execute(create_imp)

                trans.commit()

                # check the results
                df = pd.read_sql('SELECT * FROM prefix_aggregation_imputed', engine)

                # we should have a record for every entity/date combo
                assert df.shape[0] == len(states_table)

                for feat in feat_list:
                    # all of the input columns should be in the result and be null-free
                    assert 'prefix_entity_id_1y_%s_max' % feat in df.columns.values
                    assert df['prefix_entity_id_1y_%s_max' % feat].isnull().sum() == 0

                    # for non-categoricals, should add an "imputed" column and be non-null
                    # (categoricals are expected to be handled through the null category)
                    # zero_noflag imputation should not generate a flag either
                    if feat in exp_imp_cols and coltype != 'categorical' and imp != 'zero_noflag':
                        assert 'prefix_entity_id_1y_%s_max_imp' % feat in df.columns.values
                        assert df['prefix_entity_id_1y_%s_max_imp' % feat].isnull().sum() == 0
                    else:
                        # should not generate an imputed column when not needed
                        assert 'prefix_entity_id_1y_%s_max_imp' % feat not in df.columns.values



from datetime import datetime, timedelta
from unittest.mock import patch

import factory
import numpy
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition.distance_from_best import DistanceFromBestTable, BestDistancePlotter
from triage.component.catwalk.db import ensure_db

from tests.results_tests.factories import (
    EvaluationFactory,
    ModelFactory,
    ModelGroupFactory,
    init_engine,
    session,
    MatrixFactory
)

from .utils import create_sample_distance_table


def _sql_add_days(sql_date, days):
    return datetime.strftime(
        datetime.strptime(sql_date, '%Y-%m-%d') + timedelta(days=days),
        '%Y-%m-%d'
    )


def test_DistanceFromBestTable():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ensure_db(engine)
        init_engine(engine)
        model_groups = {
            'stable': ModelGroupFactory(model_type='myStableClassifier'),
            'bad': ModelGroupFactory(model_type='myBadClassifier'),
            'spiky': ModelGroupFactory(model_type='mySpikeClassifier'),
        }

        class StableModelFactory(ModelFactory):
            model_group_rel = model_groups['stable']

        class BadModelFactory(ModelFactory):
            model_group_rel = model_groups['bad']

        class SpikyModelFactory(ModelFactory):
            model_group_rel = model_groups['spiky']

        models = {
            'stable_3y_ago': StableModelFactory(train_end_time='2014-01-01'),
            'stable_2y_ago': StableModelFactory(train_end_time='2015-01-01'),
            'stable_1y_ago': StableModelFactory(train_end_time='2016-01-01'),
            'bad_3y_ago': BadModelFactory(train_end_time='2014-01-01'),
            'bad_2y_ago': BadModelFactory(train_end_time='2015-01-01'),
            'bad_1y_ago': BadModelFactory(train_end_time='2016-01-01'),
            'spiky_3y_ago': SpikyModelFactory(train_end_time='2014-01-01'),
            'spiky_2y_ago': SpikyModelFactory(train_end_time='2015-01-01'),
            'spiky_1y_ago': SpikyModelFactory(train_end_time='2016-01-01'),
        }

        class ImmediateEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(lambda o: o.model_rel.train_end_time)
            evaluation_end_time = factory.LazyAttribute(
                lambda o: _sql_add_days(o.model_rel.train_end_time, 1)
            )

        class MonthOutEvalFactory(EvaluationFactory):
            evaluation_start_time = factory.LazyAttribute(
                lambda o: _sql_add_days(o.model_rel.train_end_time, 31)
            )
            evaluation_end_time = factory.LazyAttribute(
                lambda o: _sql_add_days(o.model_rel.train_end_time, 32)
            )

        class Precision100Factory(ImmediateEvalFactory):
            metric = 'precision@'
            parameter = '100_abs'

        class Precision100FactoryMonthOut(MonthOutEvalFactory):
            metric = 'precision@'
            parameter = '100_abs'

        class Recall100Factory(ImmediateEvalFactory):
            metric = 'recall@'
            parameter = '100_abs'

        class Recall100FactoryMonthOut(MonthOutEvalFactory):
            metric = 'recall@'
            parameter = '100_abs'

        for (add_val, PrecFac, RecFac) in (
            (0, Precision100Factory, Recall100Factory),
            (-0.15, Precision100FactoryMonthOut, Recall100FactoryMonthOut)
        ):
            PrecFac(model_rel=models['stable_3y_ago'], value=0.6+add_val)
            PrecFac(model_rel=models['stable_2y_ago'], value=0.57+add_val)
            PrecFac(model_rel=models['stable_1y_ago'], value=0.59+add_val)
            PrecFac(model_rel=models['bad_3y_ago'], value=0.4+add_val)
            PrecFac(model_rel=models['bad_2y_ago'], value=0.39+add_val)
            PrecFac(model_rel=models['bad_1y_ago'], value=0.43+add_val)
            PrecFac(model_rel=models['spiky_3y_ago'], value=0.8+add_val)
            PrecFac(model_rel=models['spiky_2y_ago'], value=0.4+add_val)
            PrecFac(model_rel=models['spiky_1y_ago'], value=0.4+add_val)
            RecFac(model_rel=models['stable_3y_ago'], value=0.55+add_val)
            RecFac(model_rel=models['stable_2y_ago'], value=0.56+add_val)
            RecFac(model_rel=models['stable_1y_ago'], value=0.55+add_val)
            RecFac(model_rel=models['bad_3y_ago'], value=0.35+add_val)
            RecFac(model_rel=models['bad_2y_ago'], value=0.34+add_val)
            RecFac(model_rel=models['bad_1y_ago'], value=0.36+add_val)
            RecFac(model_rel=models['spiky_3y_ago'], value=0.35+add_val)
            RecFac(model_rel=models['spiky_2y_ago'], value=0.8+add_val)
            RecFac(model_rel=models['spiky_1y_ago'], value=0.36+add_val)
        session.commit()
        distance_table = DistanceFromBestTable(
            db_engine=engine,
            models_table='models',
            distance_table='dist_table'
        )
        metrics = [
            {'metric': 'precision@', 'parameter': '100_abs'},
            {'metric': 'recall@', 'parameter': '100_abs'}
        ]
        model_group_ids = [mg.model_group_id for mg in model_groups.values()]
        distance_table.create_and_populate(
            model_group_ids,
            ['2014-01-01', '2015-01-01', '2016-01-01'],
            metrics)

        # get an ordered list of the models/groups for a particular metric/time
        query = '''
            select model_id, raw_value, dist_from_best_case, dist_from_best_case_next_time
            from dist_table where metric = %s and parameter = %s and train_end_time = %s
            order by dist_from_best_case
        '''

        prec_3y_ago = engine.execute(query, ('precision@', '100_abs', '2014-01-01'))
        assert [row for row in prec_3y_ago] == [
            (models['spiky_3y_ago'].model_id, 0.8, 0, 0.17),
            (models['stable_3y_ago'].model_id, 0.6, 0.2, 0),
            (models['bad_3y_ago'].model_id, 0.4, 0.4, 0.18),

        ]

        recall_2y_ago = engine.execute(query, ('recall@', '100_abs', '2015-01-01'))
        assert [row for row in recall_2y_ago] == [
            (models['spiky_2y_ago'].model_id, 0.8, 0, 0.19),
            (models['stable_2y_ago'].model_id, 0.56, 0.24, 0),
            (models['bad_2y_ago'].model_id, 0.34, 0.46, 0.19),

        ]

        assert distance_table.observed_bounds == {
            ('precision@', '100_abs'): (0.39, 0.8),
            ('recall@', '100_abs'): (0.34, 0.8),
        }


def test_BestDistancePlotter():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)
        plotter = BestDistancePlotter(distance_table)
        df_dist = plotter.generate_plot_data(
            metric='precision@',
            parameter='100_abs',
            model_group_ids=[1, 2],
            train_end_times=['2014-01-01', '2015-01-01']
        )
        # assert that we have the right # of columns and a row for each % diff value
        # 202 row because 101 percentiles (0-100 inclusive), 2 model groups
        assert df_dist.shape == (101 * 2, 5)

        # all of the model groups are within .34 of the best, so pick
        # a number higher than that and all should qualify
        for value in df_dist[df_dist['distance'] == 0.35]['pct_of_time'].values:
            assert numpy.isclose(value, 1.0)

        # model group 1 (stable) should be within 0.11 1/2 of the time
        # if we included 2016 in the train_end_times, this would be 1/3!
        for value in df_dist[
            (df_dist['distance'] == 0.11) &
            (df_dist['model_group_id'] == 1)
        ]['pct_of_time'].values:
            assert numpy.isclose(value, 0.5)


def test_BestDistancePlotter_plot():
    with patch('triage.component.audition.distance_from_best.plot_cats') as plot_patch:
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            distance_table, model_groups = create_sample_distance_table(engine)
            plotter = BestDistancePlotter(distance_table)
            plotter.plot_all_best_dist(
                [{'metric': 'precision@', 'parameter': '100_abs'}],
                model_group_ids=[1, 2],
                train_end_times=['2014-01-01', '2015-01-01'],
            )
        assert plot_patch.called
        args, kwargs = plot_patch.call_args
        assert 'distance' in kwargs['frame']
        assert 'pct_of_time' in kwargs['frame']
        assert kwargs['x_col'] == 'distance'
        assert kwargs['y_col'] == 'pct_of_time'


def test_BestDistancePlotter_plot_bounds():
    class FakeDistanceTable(object):
        @property
        def observed_bounds(self):
            return {
                ('precision@', '100_abs'): (0.02, 0.87),
                ('recall@', '100_abs'): (0.0, 1.0),
                ('false positives@', '300_abs'): (2, 162)
            }

    plotter = BestDistancePlotter(FakeDistanceTable())
    assert plotter.plot_bounds('precision@', '100_abs') == (0.0, 1.0)
    assert plotter.plot_bounds('recall@', '100_abs') == (0.0, 1.0)
    assert plotter.plot_bounds('false positives@', '300_abs') == (2, 178)



from triage.component.catwalk.model_trainers import NO_FEATURE_IMPORTANCE


def _entity_feature_values(matrix, feature_name, as_of_date=None):
    """Finds the value of the given feature for each entity in a matrix

    Args:
        matrix (pandas.DataFrame), with index either 'entity_id' or 'entity_id'/'as_of_date'
        feature_name (string) The name of a column in the matrix
        as_of_date (optional) If the matrix is indexed with as_of_date,
            then this must be one of the valid as_of_dates in the matrix

    Returns: (list) of (entity_id, feature_value) tuples
    """
    if matrix.index.names == ['entity_id']:
        if feature_name == NO_FEATURE_IMPORTANCE:
            # if model does not support feature importance, write 0 as individual importance.
            return list(zip(matrix.index.values, [None] * len(matrix.index.values)))
        else:
            return list(zip(matrix.index.values, matrix[feature_name].tolist()))
    elif 'entity_id' in matrix.index.names:
        results = []
        index_of_entity = matrix.index.names.index('entity_id')
        index_of_date = matrix.index.names.index('as_of_date')

        if feature_name == NO_FEATURE_IMPORTANCE:
            zipped_iter = zip(matrix.index.values, [None] * len(matrix.index.values))
        else:
            zipped_iter = zip(matrix.index.values, matrix[feature_name].tolist())

        for row in zipped_iter:
            index_values, feature_value = row
            entity_id = index_values[index_of_entity]
            if index_values[index_of_date] == as_of_date:
                results.append((entity_id, feature_value))
        return results
    else:
        return []


def uniform_distribution(db_engine, model_id, as_of_date, test_matrix_store, n_ranks):
    """Calculates individual feature importances based on the global feature importances

    Args:
        db_engine (sqlalchemy.engine)
        model_id (int) A model id, expected to be present in model_metadata.models
        as_of_date (datetime or string) The date to produce individual importances as of
        test_matrix_store (catwalk.storage.MatrixStore) The test matrix
        n_ranks (int) Number of ranks to calculate and save. Defaults to 5

    Returns: (list) dicts with entity_id, feature_value, feature_name, score
    """
    global_feature_importances = [row for row in db_engine.execute(
        '''select feature, feature_importance
        from train_results.feature_importances where model_id = %s
        order by feature_importance desc limit %s''',
        model_id,
        n_ranks
    )]

    results = []

    for feature_name, feature_importance in global_feature_importances:
        efv = _entity_feature_values(test_matrix_store.matrix, feature_name, as_of_date)
        for entity_id, feature_value in efv:
            results.append({
                'entity_id': entity_id,
                'feature_value': feature_value,
                'feature_name': feature_name,
                'score': feature_importance,
            })
    return results



import datetime
import logging
from unittest import TestCase

from triage.util.conf import convert_str_to_relativedelta

from triage.component.timechop import Timechop

logging.basicConfig(level=logging.DEBUG)


class test_calculate_train_test_split_times(TestCase):

    def test_valid_input(self):
        expected_result = [
            datetime.datetime(2015, 3, 1, 0, 0),
            datetime.datetime(2015, 6, 1, 0, 0),
            datetime.datetime(2015, 9, 1, 0, 0),
            datetime.datetime(2015, 12, 1, 0, 0),
            datetime.datetime(2016, 3, 1, 0, 0),
            datetime.datetime(2016, 6, 1, 0, 0)
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2017, 1, 1, 0, 0),
            label_start_time=datetime.datetime(2015, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2017, 1, 1, 0, 0),
            model_update_frequency='3 months',
            training_as_of_date_frequencies=['1 day'],
            test_as_of_date_frequencies=['1 day'],
            max_training_histories=['1 year'],
            test_durations=['6 months'],
            test_label_timespans=['1 months'],
            training_label_timespans=['3 days']
        )

        # this should throw an exception because last possible label date is after
        # end of feature time
        result = chopper.calculate_train_test_split_times(
            training_label_timespan=convert_str_to_relativedelta('3 days'),
            test_duration='6 months',
            test_label_timespan=convert_str_to_relativedelta('1 month')
        )

        assert result == expected_result

    def test_labels_after_features(self):
        chopper = Timechop(
            feature_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2016, 1, 1, 0, 0),
            label_start_time=datetime.datetime(2015, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2017, 1, 1, 0, 0),
            model_update_frequency='3 months',
            training_as_of_date_frequencies=['1 day'],
            test_as_of_date_frequencies=['1 day'],
            max_training_histories=['1 year'],
            test_durations=['6 months'],
            test_label_timespans=['1 months'],
            training_label_timespans=['3 days']
        )

        # this should throw an exception because last possible label date is after
        # end of feature time
        with self.assertRaises(ValueError):
            chopper.calculate_train_test_split_times(
                training_label_timespan=convert_str_to_relativedelta('3 days'),
                test_duration='6 months',
                test_label_timespan=convert_str_to_relativedelta('1 month')
            )

    def test_no_valid_label_dates(self):
        chopper = Timechop(
            feature_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2016, 1, 1, 0, 0),
            label_start_time=datetime.datetime(2015, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2015, 2, 1, 0, 0),
            model_update_frequency='3 months',
            training_as_of_date_frequencies=['1 day'],
            test_as_of_date_frequencies=['1 day'],
            max_training_histories=['1 year'],
            test_durations=['6 months'],
            test_label_timespans=['1 months'],
            training_label_timespans=['3 days']
        )

        # this should raise an error because there are no valid label dates in
        # the labeling time (label span is longer than labeling time)
        with self.assertRaises(ValueError):
            chopper.calculate_train_test_split_times(
                training_label_timespan=convert_str_to_relativedelta('3 days'),
                test_duration='6 months',
                test_label_timespan=convert_str_to_relativedelta('1 month')
            )


def test_calculate_as_of_times_one_day_freq():
    expected_result = [
        datetime.datetime(2011, 1, 1, 0, 0),
        datetime.datetime(2011, 1, 2, 0, 0),
        datetime.datetime(2011, 1, 3, 0, 0),
        datetime.datetime(2011, 1, 4, 0, 0),
        datetime.datetime(2011, 1, 5, 0, 0),
        datetime.datetime(2011, 1, 6, 0, 0),
        datetime.datetime(2011, 1, 7, 0, 0),
        datetime.datetime(2011, 1, 8, 0, 0),
        datetime.datetime(2011, 1, 9, 0, 0),
        datetime.datetime(2011, 1, 10, 0, 0),
        datetime.datetime(2011, 1, 11, 0, 0)
    ]
    chopper = Timechop(
        feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
        feature_end_time=datetime.datetime(2012, 1, 1, 0, 0),
        label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
        label_end_time=datetime.datetime(2012, 1, 1, 0, 0),
        model_update_frequency='1 year',
        training_as_of_date_frequencies=['1 days'],
        test_as_of_date_frequencies=['7 days'],
        max_training_histories=['10 days', '1 year'],
        test_durations=['1 month'],
        test_label_timespans=['1 day'],
        training_label_timespans=['3 months']
    )
    result = chopper.calculate_as_of_times(
        as_of_start_limit=datetime.datetime(2011, 1, 1, 0, 0),
        as_of_end_limit=datetime.datetime(2011, 1, 11, 0, 0),
        data_frequency=convert_str_to_relativedelta('1 days')
    )
    assert result == expected_result


def test_calculate_as_of_times_three_day_freq():
    expected_result = [
        datetime.datetime(2011, 1, 1, 0, 0),
        datetime.datetime(2011, 1, 4, 0, 0),
        datetime.datetime(2011, 1, 7, 0, 0),
        datetime.datetime(2011, 1, 10, 0, 0),
    ]
    chopper = Timechop(
        feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
        feature_end_time=datetime.datetime(2012, 1, 1, 0, 0),
        label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
        label_end_time=datetime.datetime(2012, 1, 1, 0, 0),
        model_update_frequency='1 year',
        training_as_of_date_frequencies=['1 days'],
        test_as_of_date_frequencies=['7 days'],
        max_training_histories=['10 days', '1 year'],
        test_durations=['1 month'],
        test_label_timespans=['1 day'],
        training_label_timespans=['3 months']
    )
    result = chopper.calculate_as_of_times(
        as_of_start_limit=datetime.datetime(2011, 1, 1, 0, 0),
        as_of_end_limit=datetime.datetime(2011, 1, 11, 0, 0),
        data_frequency=convert_str_to_relativedelta('3 days'),
        forward=True
    )
    assert result == expected_result


class test_generate_matrix_definitions(TestCase):

    def test_look_back_time_equal_modeling_start(self):
        # TODO: rework this test since the test label window of 3 months
        # cannot be satisfied by the 10 day difference between modeling
        # start and end times, so it's not a very realistic case
        expected_result = {
            'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
            'label_start_time': datetime.datetime(2010, 1, 1, 0, 0),
            'feature_end_time': datetime.datetime(2010, 1, 11, 0, 0),
            'label_end_time': datetime.datetime(2010, 1, 11, 0, 0),
            'train_matrix': {
                'first_as_of_time': datetime.datetime(2010, 1, 1, 0, 0),
                'last_as_of_time': datetime.datetime(2010, 1, 5, 0, 0),
                'matrix_info_end_time': datetime.datetime(2010, 1, 6, 0, 0),
                'as_of_times': [
                    datetime.datetime(2010, 1, 1, 0, 0),
                    datetime.datetime(2010, 1, 2, 0, 0),
                    datetime.datetime(2010, 1, 3, 0, 0),
                    datetime.datetime(2010, 1, 4, 0, 0),
                    datetime.datetime(2010, 1, 5, 0, 0)
                ],
                'training_label_timespan': '1 day',
                'training_as_of_date_frequency': '1 days',
                'max_training_history': '5 days'
            },
            'test_matrices': [{
                'first_as_of_time': datetime.datetime(2010, 1, 6, 0, 0),
                'last_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                'matrix_info_end_time': datetime.datetime(2010, 1, 10, 0, 0),
                'as_of_times': [
                    datetime.datetime(2010, 1, 6, 0, 0),
                    datetime.datetime(2010, 1, 9, 0, 0)
                ],
                'test_label_timespan': '1 day',
                'test_as_of_date_frequency': '3 days',
                'test_duration': '5 days'
            }]
        }
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 11, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 11, 0, 0),
            model_update_frequency='5 days',
            training_as_of_date_frequencies=['1 days'],
            test_as_of_date_frequencies=['3 days'],
            max_training_histories=['5 days'],
            test_durations=['5 days'],
            test_label_timespans=['1 day'],
            training_label_timespans=['1 day']
        )
        result = chopper.generate_matrix_definitions(
            train_test_split_time=datetime.datetime(2010, 1, 6, 0, 0),
            training_as_of_date_frequency='1 days',
            max_training_history='5 days',
            test_duration='5 days',
            test_label_timespan='1 day',
            training_label_timespan='1 day'
        )
        assert result == expected_result

    def test_look_back_time_before_modeling_start(self):
        expected_result = {
            'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
            'label_start_time': datetime.datetime(2010, 1, 1, 0, 0),
            'feature_end_time': datetime.datetime(2010, 1, 11, 0, 0),
            'label_end_time': datetime.datetime(2010, 1, 11, 0, 0),
            'train_matrix': {
                'first_as_of_time': datetime.datetime(2010, 1, 1, 0, 0),
                'last_as_of_time': datetime.datetime(2010, 1, 5, 0, 0),
                'matrix_info_end_time': datetime.datetime(2010, 1, 6, 0, 0),
                'as_of_times': [
                    datetime.datetime(2010, 1, 1, 0, 0),
                    datetime.datetime(2010, 1, 2, 0, 0),
                    datetime.datetime(2010, 1, 3, 0, 0),
                    datetime.datetime(2010, 1, 4, 0, 0),
                    datetime.datetime(2010, 1, 5, 0, 0)
                ],
                'training_label_timespan': '1 day',
                'training_as_of_date_frequency': '1 days',
                'max_training_history': '10 days'
            },
            'test_matrices': [
                {
                    'first_as_of_time': datetime.datetime(2010, 1, 6, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0)
                    ],
                    'test_label_timespan': '1 day',
                    'test_as_of_date_frequency': '3 days',
                    'test_duration': '5 days'
                },
                {
                    'first_as_of_time': datetime.datetime(2010, 1, 6, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 6, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 7, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 6, 0, 0),
                    ],
                    'test_label_timespan': '1 day',
                    'test_as_of_date_frequency': '6 days',
                    'test_duration': '5 days'
                }
            ]
        }
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 11, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 11, 0, 0),
            model_update_frequency='5 days',
            training_as_of_date_frequencies=['1 days'],
            test_as_of_date_frequencies=['3 days', '6 days'],
            max_training_histories=['10 days'],
            test_durations=['5 days'],
            test_label_timespans=['1 day'],
            training_label_timespans=['1 day']
        )
        result = chopper.generate_matrix_definitions(
            train_test_split_time=datetime.datetime(2010, 1, 6, 0, 0),
            training_as_of_date_frequency='1 days',
            max_training_history='10 days',
            test_duration='5 days',
            test_label_timespan='1 day',
            training_label_timespan='1 day'
        )
        assert result == expected_result


class test_chop_time(TestCase):
    def test_evenly_divisible_values(self):
        expected_result = [
            {
                'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
                'label_start_time': datetime.datetime(2010, 1, 1, 0, 0),
                'feature_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'label_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'train_matrix': {
                    'first_as_of_time': datetime.datetime(2010, 1, 1, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 4, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 5, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 1, 0, 0),
                        datetime.datetime(2010, 1, 2, 0, 0),
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0)
                    ],
                    'training_label_timespan': '1 day',
                    'training_as_of_date_frequency': '1 days',
                    'max_training_history': '5 days'
                },
                'test_matrices': [{
                    'first_as_of_time': datetime.datetime(2010, 1, 5, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0)
                    ],
                    'test_label_timespan': '1 day',
                    'test_as_of_date_frequency': '1 days',
                    'test_duration': '5 days'
                }]
            },
            {
                'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
                'label_start_time': datetime.datetime(2010, 1, 1, 0, 0),
                'feature_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'label_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'train_matrix': {
                    'first_as_of_time': datetime.datetime(2010, 1, 4, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 4, 0, 0),
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0)
                    ],
                    'training_label_timespan': '1 day',
                    'training_as_of_date_frequency': '1 days',
                    'max_training_history': '5 days'
                },
                'test_matrices': [{
                    'first_as_of_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 14, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 15, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 10, 0, 0),
                        datetime.datetime(2010, 1, 11, 0, 0),
                        datetime.datetime(2010, 1, 12, 0, 0),
                        datetime.datetime(2010, 1, 13, 0, 0),
                        datetime.datetime(2010, 1, 14, 0, 0)
                    ],
                    'test_label_timespan': '1 day',
                    'test_as_of_date_frequency': '1 days',
                    'test_duration': '5 days'
                }]
            }
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            model_update_frequency='5 days',
            training_as_of_date_frequencies=['1 days'],
            test_as_of_date_frequencies=['1 days'],
            max_training_histories=['5 days'],
            test_durations=['5 days'],
            test_label_timespans=['1 day'],
            training_label_timespans=['1 day']
        )
        result = chopper.chop_time()
        assert(result == expected_result)

    def test_training_label_timespan_longer_than_1_day(self):
        expected_result = [
            {
                'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
                'label_start_time': datetime.datetime(2010, 1, 1, 0, 0),
                'feature_end_time': datetime.datetime(2010, 1, 19, 0, 0),
                'label_end_time': datetime.datetime(2010, 1, 19, 0, 0),
                'train_matrix': {
                    'first_as_of_time': datetime.datetime(2010, 1, 1, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 4, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 1, 0, 0),
                        datetime.datetime(2010, 1, 2, 0, 0),
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0)
                    ],
                    'training_label_timespan': '5 days',
                    'training_as_of_date_frequency': '1 days',
                    'max_training_history': '5 days'
                },
                'test_matrices': [{
                    'first_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 13, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 18, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 9, 0, 0),
                        datetime.datetime(2010, 1, 10, 0, 0),
                        datetime.datetime(2010, 1, 11, 0, 0),
                        datetime.datetime(2010, 1, 12, 0, 0),
                        datetime.datetime(2010, 1, 13, 0, 0)
                    ],
                    'test_label_timespan': '5 days',
                    'test_as_of_date_frequency': '1 days',
                    'test_duration': '5 days'
                }]
            }
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 19, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 19, 0, 0),
            model_update_frequency='5 days',
            training_as_of_date_frequencies=['1 days'],
            test_as_of_date_frequencies=['1 days'],
            max_training_histories=['5 days'],
            test_durations=['5 days'],
            test_label_timespans=['5 days'],
            training_label_timespans=['5 days']
        )
        result = chopper.chop_time()
        assert(result == expected_result)

    def test_unevenly_divisible_lookback_duration(self):
        expected_result = [
            {
                'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
                'label_start_time': datetime.datetime(2010, 1, 1, 0, 0),
                'feature_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'label_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'train_matrix': {
                    'first_as_of_time': datetime.datetime(2010, 1, 1, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 4, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 5, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 1, 0, 0),
                        datetime.datetime(2010, 1, 2, 0, 0),
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0)
                    ],
                    'training_label_timespan': '1 day',
                    'training_as_of_date_frequency': '1 days',
                    'max_training_history': '7 days'
                },
                'test_matrices': [{
                    'first_as_of_time': datetime.datetime(2010, 1, 5, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0)
                    ],
                    'test_label_timespan': '1 day',
                    'test_as_of_date_frequency': '1 days',
                    'test_duration': '5 days'
                }]
            },
            {
                'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
                'label_start_time': datetime.datetime(2010, 1, 1, 0, 0),
                'feature_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'label_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'train_matrix': {
                    'first_as_of_time': datetime.datetime(2010, 1, 2, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 2, 0, 0),
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0),
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0)
                    ],
                    'training_label_timespan': '1 day',
                    'training_as_of_date_frequency': '1 days',
                    'max_training_history': '7 days'
                },
                'test_matrices': [{
                    'first_as_of_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 14, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 15, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 10, 0, 0),
                        datetime.datetime(2010, 1, 11, 0, 0),
                        datetime.datetime(2010, 1, 12, 0, 0),
                        datetime.datetime(2010, 1, 13, 0, 0),
                        datetime.datetime(2010, 1, 14, 0, 0)
                    ],
                    'test_label_timespan': '1 day',
                    'test_as_of_date_frequency': '1 days',
                    'test_duration': '5 days'
                }]
            }
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 1, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            model_update_frequency='5 days',
            training_as_of_date_frequencies=['1 days'],
            test_as_of_date_frequencies=['1 days'],
            max_training_histories=['7 days'],
            test_durations=['5 days'],
            test_label_timespans=['1 day'],
            training_label_timespans=['1 day']
        )
        result = chopper.chop_time()
        assert(result == expected_result)

    def test_unevenly_divisible_update_window(self):
        expected_result = [
            {
                'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
                'label_start_time': datetime.datetime(2010, 1, 3, 0, 0),
                'feature_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'label_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'train_matrix': {
                    'first_as_of_time': datetime.datetime(2010, 1, 3, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 4, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 5, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 3, 0, 0),
                        datetime.datetime(2010, 1, 4, 0, 0)
                    ],
                    'training_label_timespan': '1 day',
                    'training_as_of_date_frequency': '1 days',
                    'max_training_history': '5 days'
                },
                'test_matrices': [{
                    'first_as_of_time': datetime.datetime(2010, 1, 5, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0)
                    ],
                    'test_label_timespan': '1 day',
                    'test_as_of_date_frequency': '1 days',
                    'test_duration': '5 days'
                }]
            },
            {
                'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
                'label_start_time': datetime.datetime(2010, 1, 3, 0, 0),
                'feature_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'label_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'train_matrix': {
                    'first_as_of_time': datetime.datetime(2010, 1, 4, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 9, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 4, 0, 0),
                        datetime.datetime(2010, 1, 5, 0, 0),
                        datetime.datetime(2010, 1, 6, 0, 0),
                        datetime.datetime(2010, 1, 7, 0, 0),
                        datetime.datetime(2010, 1, 8, 0, 0),
                        datetime.datetime(2010, 1, 9, 0, 0)
                    ],
                    'training_label_timespan': '1 day',
                    'training_as_of_date_frequency': '1 days',
                    'max_training_history': '5 days'
                },
                'test_matrices': [{
                    'first_as_of_time': datetime.datetime(2010, 1, 10, 0, 0),
                    'last_as_of_time': datetime.datetime(2010, 1, 14, 0, 0),
                    'matrix_info_end_time': datetime.datetime(2010, 1, 15, 0, 0),
                    'as_of_times': [
                        datetime.datetime(2010, 1, 10, 0, 0),
                        datetime.datetime(2010, 1, 11, 0, 0),
                        datetime.datetime(2010, 1, 12, 0, 0),
                        datetime.datetime(2010, 1, 13, 0, 0),
                        datetime.datetime(2010, 1, 14, 0, 0)
                    ],
                    'test_label_timespan': '1 day',
                    'test_as_of_date_frequency': '1 days',
                    'test_duration': '5 days'
                }]
            }
        ]
        chopper = Timechop(
            feature_start_time=datetime.datetime(1990, 1, 1, 0, 0),
            feature_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            label_start_time=datetime.datetime(2010, 1, 3, 0, 0),
            label_end_time=datetime.datetime(2010, 1, 16, 0, 0),
            model_update_frequency='5 days',
            training_as_of_date_frequencies=['1 days'],
            test_as_of_date_frequencies=['1 days'],
            max_training_histories=['5 days'],
            test_durations=['5 days'],
            test_label_timespans=['1 day'],
            training_label_timespans=['1 day']
        )
        result = chopper.chop_time()
        assert(result == expected_result)


class test__init__(TestCase):
    def test_bad_feature_start_time(self):
        with self.assertRaises(ValueError):
            Timechop(
                feature_start_time=datetime.datetime(2011, 1, 1, 0, 0),
                feature_end_time=datetime.datetime(2010, 1, 16, 0, 0),
                label_start_time=datetime.datetime(2010, 1, 3, 0, 0),
                label_end_time=datetime.datetime(2010, 1, 16, 0, 0),
                model_update_frequency='5 days',
                training_as_of_date_frequencies=['1 days'],
                test_as_of_date_frequencies=['1 days'],
                max_training_histories=['5 days'],
                test_durations=['5 days'],
                test_label_timespans=['1 day'],
                training_label_timespans=['1 day']
            )



from datetime import date, timedelta

import testing.postgresql
from sqlalchemy import create_engine

from triage.component.architect.label_generators import LabelGenerator

from .utils import create_binary_outcome_events

events_data = [
    # entity id, event_date, outcome
    [1, date(2014, 1, 1), True],
    [1, date(2014, 11, 10), False],
    [1, date(2015, 1, 1), False],
    [1, date(2015, 11, 10), True],
    [2, date(2014, 6, 8), True],
    [2, date(2015, 6, 8), False],
    [3, date(2014, 3, 3), False],
    [3, date(2014, 7, 24), False],
    [3, date(2015, 3, 3), True],
    [3, date(2015, 7, 24), False],
    [4, date(2014, 12, 13), False],
    [4, date(2015, 12, 13), False],
]


def test_label_generation():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_binary_outcome_events(engine, 'events', events_data)

        labels_table_name = 'labels'
        query = """select
            events.entity_id,
            bool_or(outcome::bool)::integer as outcome
        from events
        where
            '{as_of_date}' <= outcome_date
            and outcome_date < '{as_of_date}'::timestamp + interval '{label_timespan}'
            group by entity_id
        """
        label_generator = LabelGenerator(db_engine=engine, query=query)
        label_generator._create_labels_table(labels_table_name)
        label_generator.generate(
            start_date='2014-09-30',
            label_timespan='6months',
            labels_table='labels'
        )

        expected = [
            # entity_id, as_of_date, label_timespan, name, type, label
            (1, date(2014, 9, 30), timedelta(180), 'outcome', 'binary', False),
            (3, date(2014, 9, 30), timedelta(180), 'outcome', 'binary', True),
            (4, date(2014, 9, 30), timedelta(180), 'outcome', 'binary', False),
        ]
        result = engine.execute(
            'select * from {} order by entity_id, as_of_date'.format(
                labels_table_name)
        )
        records = [row for row in result]
        assert records == expected


def test_generate_all_labels():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_binary_outcome_events(engine, 'events', events_data)

        labels_table_name = 'labels'

        query = """select
            events.entity_id,
            bool_or(outcome::bool)::integer as outcome
        from events
        where
            '{as_of_date}' <= outcome_date
            and outcome_date < '{as_of_date}'::timestamp + interval '{label_timespan}'
            group by entity_id
        """

        label_generator = LabelGenerator(db_engine=engine, query=query)
        label_generator.generate_all_labels(
            labels_table=labels_table_name,
            as_of_dates=['2014-09-30', '2015-03-30'],
            label_timespans=['6month', '3month'],
        )

        result = engine.execute('''
            select * from {}
            order by entity_id, as_of_date, label_timespan desc
        '''.format(labels_table_name)
        )
        records = [row for row in result]

        expected = [
            # entity_id, as_of_date, label_timespan, name, type, label
            (1, date(2014, 9, 30), timedelta(180), 'outcome', 'binary', False),
            (1, date(2014, 9, 30), timedelta(90), 'outcome', 'binary', False),
            (2, date(2015, 3, 30), timedelta(180), 'outcome', 'binary', False),
            (2, date(2015, 3, 30), timedelta(90), 'outcome', 'binary', False),
            (3, date(2014, 9, 30), timedelta(180), 'outcome', 'binary', True),
            (3, date(2015, 3, 30), timedelta(180), 'outcome', 'binary', False),
            (4, date(2014, 9, 30), timedelta(180), 'outcome', 'binary', False),
            (4, date(2014, 9, 30), timedelta(90), 'outcome', 'binary', False),
        ]
        assert records == expected



import re
import warnings
from dateutil.relativedelta import relativedelta
from datetime import datetime


def dt_from_str(dt_str):
    if isinstance(dt_str, datetime):
        return dt_str
    return datetime.strptime(dt_str, '%Y-%m-%d')


def parse_delta_string(delta_string):
    """Given a string in a postgres interval format (e.g., '1 month'),
    parse the units and value from it.

    Assumptions:
    - The string is in the format 'value unit', where
      value is an int and unit is one of year(s), month(s), day(s),
      week(s), hour(s), minute(s), second(s), microsecond(s), or an
      abbreviation matching y, d, w, h, m, s, or ms (case-insensitive).
      For example: 1 year, 1year, 2 years, 1 y, 2y, 1Y.

    :param delta_string: the time interval to convert
    :type delta_string: str

    :return: time units, number of units (value)
    :rtype: tuple

    :raises: ValueError if the delta_string is not in the expected format

    """
    match = parse_delta_string.pattern.search(delta_string)
    if match:
        (pre_value, units) = match.groups()
        return (units, int(pre_value))

    raise ValueError('Could not parse value from time delta string: {!r}'
                     .format(delta_string))


parse_delta_string.pattern = re.compile(r'^(\d+) *([^ ]+)$')


def convert_str_to_relativedelta(delta_string):
    """Given a string in a postgres interval format (e.g., '1 month'),
    convert it to a dateutil.relativedelta.relativedelta.

    Assumptions:
    - The string is in the format 'value unit', where
      value is an int and unit is one of year(s), month(s), day(s),
      week(s), hour(s), minute(s), second(s), microsecond(s), or an
      abbreviation matching y, d, w, h, m, s, or ms (case-insensitive).
      For example: 1 year, 1year, 2 years, 1 y, 2y, 1Y.

    :param delta_string: the time interval to convert
    :type delta_string: str

    :return: the time interval as a relativedelta
    :rtype: dateutil.relativedelta.relativedelta

    :raises: ValueError if the delta_string is not in the expected format

    """
    (units, value) = parse_delta_string(delta_string)

    verbose_match = convert_str_to_relativedelta.pattern_verbose.search(units)
    if verbose_match:
        unit_type = verbose_match.group(1) + 's'
        return relativedelta(**{unit_type: value})

    try:
        unit_type = convert_str_to_relativedelta.brief_units[units.lower()]
    except KeyError:
        pass
    else:
        if unit_type == 'minutes':
            warnings.warn(
                'Time delta units "{}" converted to minutes.'.format(units),
                RuntimeWarning
            )
        return relativedelta(**{unit_type: value})

    raise ValueError('Could not handle units. Units: {} Value: {}'
                     .format(units, value))


convert_str_to_relativedelta.pattern_verbose = re.compile(
    r'^(year|month|day|week|hour|minute|second|microsecond)s?$'
)

convert_str_to_relativedelta.brief_units = {
    'y': 'years',
    'd': 'days',
    'w': 'weeks',
    'h': 'hours',
    'm': 'minutes',
    's': 'seconds',
    'ms': 'microseconds',
}



from triage.component.catwalk.utils import filename_friendly_hash, \
    save_experiment_and_get_hash, \
    sort_predictions_and_labels
from triage.component.catwalk.db import ensure_db
from sqlalchemy import create_engine
import testing.postgresql
import datetime
import logging
import re


def test_filename_friendly_hash():
    data = {
        'stuff': 'stuff',
        'other_stuff': 'more_stuff',
        'a_datetime': datetime.datetime(2015, 1, 1),
        'a_date': datetime.date(2016, 1, 1),
        'a_number': 5.0
    }
    output = filename_friendly_hash(data)
    assert isinstance(output, str)
    assert re.match('^[\w]+$', output) is not None

    # make sure ordering keys differently doesn't change the hash
    new_output = filename_friendly_hash({
        'other_stuff': 'more_stuff',
        'stuff': 'stuff',
        'a_datetime': datetime.datetime(2015, 1, 1),
        'a_date': datetime.date(2016, 1, 1),
        'a_number': 5.0
    })
    assert new_output == output

    # make sure new data hashes to something different
    new_output = filename_friendly_hash({
        'stuff': 'stuff',
        'a_number': 5.0
    })
    assert new_output != output


def test_filename_friendly_hash_stability():
    nested_data = {
        'one': 'two',
        'three': {
            'four': 'five',
            'six': 'seven'
        }
    }
    output = filename_friendly_hash(nested_data)
    # 1. we want to make sure this is stable across different runs
    # so hardcode an expected value
    assert output == '9a844a7ebbfd821010b1c2c13f7391e6'
    other_nested_data = {
        'one': 'two',
        'three': {
            'six': 'seven',
            'four': 'five'
        }
    }
    new_output = filename_friendly_hash(other_nested_data)
    assert output == new_output


def test_save_experiment_and_get_hash():
    # no reason to make assertions on the config itself, use a basic dict
    experiment_config = {'one': 'two'}
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        ensure_db(engine)
        exp_hash = save_experiment_and_get_hash(experiment_config, engine)
        assert isinstance(exp_hash, str)
        new_hash = save_experiment_and_get_hash(experiment_config, engine)
        assert new_hash == exp_hash

def test_sort_predictions_and_labels():
    predictions = [
        0.5,
        0.4,
        0.6,
        0.5,
    ]

    labels = [
        False,
        False,
        True,
        True
    ]

    sorted_predictions, sorted_labels = sort_predictions_and_labels(
        predictions,
        labels,
        8
    )
    assert sorted_predictions == (0.6, 0.5, 0.5, 0.4)
    assert sorted_labels == (True, True, False, False)


    sorted_predictions, sorted_labels = sort_predictions_and_labels(
        predictions,
        labels,
        12345
    )
    assert sorted_predictions == (0.6, 0.5, 0.5, 0.4)
    assert sorted_labels == (True, False, True, False)



"""Functions to retrieve basic information about tables in a Postgres database"""
from sqlalchemy import MetaData, Table


def split_table(table_name):
    """Split a fully-qualified table name into schema and table

    Args:
        table_name (string) A table name, either with or without a schema prefix

    Returns: (tuple) of schema and table name
    """
    table_parts = table_name.split('.')
    if len(table_parts) == 2:
        return tuple(table_parts)
    elif len(table_parts) == 1:
        return (None, table_parts[0])
    else:
        raise ValueError('Table name in unknown format')


def table_object(table_name, db_engine):
    """Produce a table object for the given table name

    This does not load data about the table from the engine yet,
    so it is safe to call for a table that doesn't exist.

    Args:
        table_name (string) A table name (with schema)
        db_engine (sqlalchemy.engine)

    Returns: (sqlalchemy.Table)
    """
    schema, table = split_table(table_name)
    meta = MetaData(schema=schema, bind=db_engine)
    return Table(table, meta)


def reflected_table(table_name, db_engine):
    """Produce a loaded table object for the given table name

    Will attempt to load the metadata about the table from the database
    So this will fail if the table doesn't exist.

    Args:
        table_name (string) A table name (with schema)
        db_engine (sqlalchemy.engine)

    Returns: (sqlalchemy.Table) A loaded table object
    """
    schema, table = split_table(table_name)
    meta = MetaData(schema=schema, bind=db_engine)
    return Table(table, meta, autoload=True, autoload_from=db_engine)


def table_exists(table_name, db_engine):
    """Checks whether the table exists

    Args:
        table_name (string) A table name (with schema)
        db_engine (sqlalchemy.engine)

    Returns: (boolean) Whether or not the table exists in the database
    """
    return table_object(table_name, db_engine).exists()


def table_has_data(table_name, db_engine):
    """Check whether the table contains any data

    Args:
        table_name (string) A table name (with schema)
        db_engine (sqlalchemy.engine)

    Returns: (boolean) Whether or not the table has any data
    """
    if not table_exists(table_name, db_engine):
        return False
    results = [
        row for row in
        db_engine.execute('select * from {} limit 1'.format(table_name))
    ]

    return len(results) > 0


def table_has_column(table_name, column, db_engine):
    """Check whether the table contains a column of the given name

    The table is expected to exist.

    Args:
        table_name (string) A table name (with schema)
        column (string) A column name
        db_engine (sqlalchemy.engine)

    Returns: (boolean) Whether or not the table contains the column
    """
    return column in reflected_table(table_name, db_engine).columns


def column_type(table_name, column, db_engine):
    """Find the database type of the given column in the given table

    The table is expected to exist, and contain a column of the given name

    Args:
        table_name (string) A table name (with schema)
        column (string) A column name
        db_engine (sqlalchemy.engine)

    Returns: (sqlalchemy.types) The DDL type of the column; For instance,
        sqlalchemy.types.BOOLEAN instead of
        sqlalchemy.types.Boolean
    """
    return type(reflected_table(table_name, db_engine).columns[column].type)
